#!/usr/bin/python
#
# Copyright 2017 Red Hat | Ansible, Alex Grönholm <alex.gronholm@nextday.fi>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: docker_volume
short_description: Manage Docker volumes
description:
  - Create/remove Docker volumes.
  - Performs largely the same function as the C(docker volume) CLI subcommand.
extends_documentation_fragment:
  - community.docker._docker.api_documentation
  - community.docker._attributes
  - community.docker._attributes.actiongroup_docker

attributes:
  check_mode:
    support: full
  diff_mode:
    support: full
  idempotent:
    support: partial
    details:
      - If O(recreate=always) the module is not idempotent.

options:
  volume_name:
    description:
      - Name of the volume to operate on.
    type: str
    required: true
    aliases:
      - name

  driver:
    description:
      - Specify the type of volume. Docker provides the V(local) driver, but 3rd party drivers can also be used.
    type: str
    default: local

  driver_options:
    description:
      - 'Dictionary of volume settings. Consult the Docker documentation for valid options and values:
        U(https://docs.docker.com/engine/reference/commandline/volume_create/#driver-specific-options).'
    type: dict
    default: {}

  labels:
    description:
      - Dictionary of label key/values to set for the volume.
    type: dict

  recreate:
    description:
      - Controls when a volume will be recreated when O(state=present). Please note that recreating an existing volume will
        cause B(any data in the existing volume to be lost!) The volume will be deleted and a new volume with the same name
        will be created.
      - The value V(always) forces the volume to be always recreated.
      - The value V(never) makes sure the volume will not be recreated.
      - The value V(options-changed) makes sure the volume will be recreated if the volume already exist and the driver, driver
        options or labels differ.
    type: str
    default: never
    choices:
      - always
      - never
      - options-changed

  state:
    description:
      - V(absent) deletes the volume.
      - V(present) creates the volume, if it does not already exist.
    type: str
    default: present
    choices:
      - absent
      - present

author:
  - Alex Grönholm (@agronholm)

requirements:
  - "Docker API >= 1.25"
"""

EXAMPLES = r"""
---
- name: Create a volume
  community.docker.docker_volume:
    name: volume_one

- name: Remove a volume
  community.docker.docker_volume:
    name: volume_one
    state: absent

- name: Create a volume with options
  community.docker.docker_volume:
    name: volume_two
    driver_options:
      type: btrfs
      device: /dev/sda2
"""

RETURN = r"""
volume:
  description:
    - Volume inspection results for the affected volume.
  returned: success
  type: dict
  sample: {}
"""

import traceback
import typing as t

from ansible.module_utils.common.text.converters import to_text

from ansible_collections.community.docker.plugins.module_utils._api.errors import (
    APIError,
    DockerException,
)
from ansible_collections.community.docker.plugins.module_utils._common_api import (
    AnsibleDockerClient,
    RequestException,
)
from ansible_collections.community.docker.plugins.module_utils._util import (
    DifferenceTracker,
    DockerBaseClass,
    sanitize_labels,
)


class TaskParameters(DockerBaseClass):
    volume_name: str

    def __init__(self, client: AnsibleDockerClient) -> None:
        super().__init__()
        self.client = client

        self.driver: str = "local"
        self.driver_options: dict[str, t.Any] = {}
        self.labels: dict[str, t.Any] | None = None
        self.recreate: t.Literal["always", "never", "options-changed"] = "never"
        self.debug: bool = False
        self.state: t.Literal["present", "absent"] = "present"

        for key, value in client.module.params.items():
            setattr(self, key, value)


class DockerVolumeManager:
    def __init__(self, client: AnsibleDockerClient) -> None:
        self.client = client
        self.parameters = TaskParameters(client)
        self.check_mode = self.client.check_mode
        self.actions: list[str] = []
        self.results: dict[str, t.Any] = {"changed": False, "actions": self.actions}
        self.diff = self.client.module._diff
        self.diff_tracker = DifferenceTracker()
        self.diff_result: dict[str, t.Any] = {}

        self.existing_volume = self.get_existing_volume()

        state = self.parameters.state
        if state == "present":
            self.present()
        elif state == "absent":
            self.absent()

        if self.diff or self.check_mode or self.parameters.debug:
            if self.diff:
                self.diff_result["before"], self.diff_result["after"] = (
                    self.diff_tracker.get_before_after()
                )
            self.results["diff"] = self.diff_result

    def get_existing_volume(self) -> dict[str, t.Any] | None:
        try:
            volumes = self.client.get_json("/volumes")
        except APIError as e:
            self.client.fail(to_text(e))

        if volumes["Volumes"] is None:
            return None

        for volume in volumes["Volumes"]:
            if volume["Name"] == self.parameters.volume_name:
                return volume

        return None

    def has_different_config(self) -> DifferenceTracker:
        """
        Return the list of differences between the current parameters and the existing volume.

        :return: list of options that differ
        """
        assert self.existing_volume is not None
        differences = DifferenceTracker()
        if (
            self.parameters.driver
            and self.parameters.driver != self.existing_volume["Driver"]
        ):
            differences.add(
                "driver",
                parameter=self.parameters.driver,
                active=self.existing_volume["Driver"],
            )
        if self.parameters.driver_options:
            if not self.existing_volume.get("Options"):
                differences.add(
                    "driver_options",
                    parameter=self.parameters.driver_options,
                    active=self.existing_volume.get("Options"),
                )
            else:
                for key, value in self.parameters.driver_options.items():
                    if (
                        not self.existing_volume["Options"].get(key)
                        or value != self.existing_volume["Options"][key]
                    ):
                        differences.add(
                            f"driver_options.{key}",
                            parameter=value,
                            active=self.existing_volume["Options"].get(key),
                        )
        if self.parameters.labels:
            existing_labels = self.existing_volume.get("Labels") or {}
            for label in self.parameters.labels:
                if existing_labels.get(label) != self.parameters.labels.get(label):
                    differences.add(
                        f"labels.{label}",
                        parameter=self.parameters.labels.get(label),
                        active=existing_labels.get(label),
                    )

        return differences

    def create_volume(self) -> None:
        if not self.existing_volume:
            if not self.check_mode:
                try:
                    data = {
                        "Name": self.parameters.volume_name,
                        "Driver": self.parameters.driver,
                        "DriverOpts": self.parameters.driver_options,
                    }
                    if self.parameters.labels is not None:
                        data["Labels"] = self.parameters.labels
                    resp = self.client.post_json_to_json("/volumes/create", data=data)
                    self.existing_volume = self.client.get_json(
                        "/volumes/{0}", resp["Name"]
                    )
                except APIError as e:
                    self.client.fail(to_text(e))

            self.actions.append(
                f"Created volume {self.parameters.volume_name} with driver {self.parameters.driver}"
            )
            self.results["changed"] = True

    def remove_volume(self) -> None:
        if self.existing_volume:
            if not self.check_mode:
                try:
                    self.client.delete_call("/volumes/{0}", self.parameters.volume_name)
                except APIError as e:
                    self.client.fail(to_text(e))

            self.actions.append(f"Removed volume {self.parameters.volume_name}")
            self.results["changed"] = True

    def present(self) -> None:
        differences = DifferenceTracker()
        if self.existing_volume:
            differences = self.has_different_config()

        self.diff_tracker.add(
            "exists", parameter=True, active=self.existing_volume is not None
        )
        if (
            not differences.empty and self.parameters.recreate == "options-changed"
        ) or self.parameters.recreate == "always":
            self.remove_volume()
            self.existing_volume = None

        self.create_volume()

        if self.diff or self.check_mode or self.parameters.debug:
            self.diff_result["differences"] = differences.get_legacy_docker_diffs()
            self.diff_tracker.merge(differences)

        if not self.check_mode and not self.parameters.debug:
            self.results.pop("actions")

        volume_facts = self.get_existing_volume()
        self.results["volume"] = volume_facts

    def absent(self) -> None:
        self.diff_tracker.add(
            "exists", parameter=False, active=self.existing_volume is not None
        )
        self.remove_volume()


def main() -> None:
    argument_spec = {
        "volume_name": {"type": "str", "required": True, "aliases": ["name"]},
        "state": {
            "type": "str",
            "default": "present",
            "choices": ["present", "absent"],
        },
        "driver": {"type": "str", "default": "local"},
        "driver_options": {"type": "dict", "default": {}},
        "labels": {"type": "dict"},
        "recreate": {
            "type": "str",
            "default": "never",
            "choices": ["always", "never", "options-changed"],
        },
        "debug": {"type": "bool", "default": False},
    }

    client = AnsibleDockerClient(
        argument_spec=argument_spec,
        supports_check_mode=True,
        # "The docker server >= 1.9.0"
    )
    sanitize_labels(client.module.params["labels"], "labels", client)

    try:
        cm = DockerVolumeManager(client)
        client.module.exit_json(**cm.results)
    except DockerException as e:
        client.fail(
            f"An unexpected Docker error occurred: {e}",
            exception=traceback.format_exc(),
        )
    except RequestException as e:
        client.fail(
            f"An unexpected requests error occurred when trying to talk to the Docker daemon: {e}",
            exception=traceback.format_exc(),
        )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 2024 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The Vrf_global parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


def _tmplt_ip_route(route_item):
    command = "ip route {source} {destination}".format(**route_item)
    if route_item.get("tags"):
        tag_item = route_item.get("tags")
        command += " tag {tag_value}".format(**tag_item)
        if route_item.get("tags").get("route_pref"):
            command += " {route_pref}".format(**tag_item)
    if route_item.get("vrf"):
        command += " vrf {vrf}".format(**route_item)
    if route_item.get("track"):
        command += " track {track}".format(**route_item)
    return command


class Vrf_globalTemplate(NetworkTemplate):
    def __init__(self, lines=None, module=None):
        super(Vrf_globalTemplate, self).__init__(lines=lines, tmplt=self, module=module)

    # fmt: off
    PARSERS = [
        {
            "name": "name",
            "getval": re.compile(
                r"""
                ^vrf\scontext\s(?P<name>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "vrf context {{ name }}",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                    },
                },
            },
            "shared": True,
        },
        {
            "name": "description",
            "getval": re.compile(
                r"""
                \s+description\s(?P<description>.+$)
                $""", re.VERBOSE,
            ),
            "setval": "description {{ description }}",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        'description': '{{ description }}',
                    },
                },
            },
        },
        {
            "name": "rd",
            "getval": re.compile(
                r"""
                \s+rd\s(?P<rd>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "rd {{ rd }}",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        'rd': '{{ rd }}',
                    },
                },
            },
        },
        {
            "name": "ip.auto_discard",
            "getval": re.compile(
                r"""
                \s+ip\s(?P<auto_disc>auto-discard)
                $""", re.VERBOSE,
            ),
            "setval": "ip auto-discard",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "auto_discard": "{{ True if auto_disc is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "ip.domain_list",
            "getval": re.compile(
                r"""
                \s+ip\sdomain-list\s(?P<domain_list>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "ip domain-list {{ domain_list }}",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "domain_list": [
                                "{{ domain_list }}",
                            ],
                        },
                    },
                },
            },
        },
        {
            "name": "ip.domain_name",
            "getval": re.compile(
                r"""
                \s+ip\sdomain-name\s(?P<domain_name>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "ip domain-name {{ ip.domain_name }}",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "domain_name": "{{ domain_name }}",
                        },
                    },
                },
            },
        },
        {
            "name": "ip.icmp_err.source_interface",
            "getval": re.compile(
                r"""
                \s+ip\sicmp-errors
                \ssource-interface\s(?P<interface>eth|po|lo)
                (?P<interface_val>(\d+\S*))
                $""", re.VERBOSE,
            ),
            "setval": "ip icmp-errors source-interface {{ ip.icmp_err.source_interface.interface }} {{ ip.icmp_err.source_interface.interface_value }}",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "icmp_err": {
                                "source_interface": {
                                    "interface": "{{ 'ethernet' if 'eth' in interface }}"
                                    "{{ 'port-channel' if 'po' in interface }}"
                                    "{{ 'loopback' if 'lo' in interface }}",
                                    "interface_value": "{{ interface_val }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip.igmp.ssm_translate",
            "getval": re.compile(
                r"""
                \s+ip\sigmp
                \sssm-translate
                \s(?P<group_val>\S+)
                \s(?P<source_val>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "ip igmp ssm-translate {{ group }} {{ source }}",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "igmp": {
                                "ssm_translate": [
                                    {
                                        "group": "{{ group_val }}",
                                        "source": "{{ source_val }}",
                                    },
                                ],
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip.mroutes",
            "getval": re.compile(
                r"""
                \s+ip\smroute
                \s(?P<group_val>\S+)
                \s(?P<source_val>\S+)
                (\s(?P<pref_val>\d+))?
                (\svrf\s(?P<vrf_val>\S+))?
                $""", re.VERBOSE,
            ),
            "setval": "ip mroute {{ group }} {{ source }}"
            "{{ ' ' + preference|string if preference is defined }}"
            "{{ ' vrf ' + vrf if vrf is defined }}",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "mroutes": [
                                {
                                    "group": "{{ group_val }}",
                                    "source": "{{ source_val }}",
                                    "preference": "{{ pref_val if pref_val is defined }}",
                                    "vrf": "{{ vrf_val if vrf_val is defined }}",
                                },
                            ],
                        },
                    },
                },
            },
        },
        {
            "name": "ip.multicast.group_range_prefix_list",
            "getval": re.compile(
                r"""
                \s+ip\smulticast
                \sprefix-list\s(?P<prefix_lst>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "ip multicast group-range prefix-list {{ ip.multicast.group_range_prefix_list.group_range_prefix_list }}",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "multicast": {
                                "group_range_prefix_list": "{{ prefix_lst }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip.multicast.multipath.resilient",
            "getval": re.compile(
                r"""
                \s+ip\smulticast
                \smultipath\s(?P<res>resilient)
                $""", re.VERBOSE,
            ),
            "setval": "ip multicast multipath resilient",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "multicast": {
                                "multipath": {
                                    "resilient": "{{ True if res is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip.multicast.multipath.splitting_type.none",
            "getval": re.compile(
                r"""
                \s+ip\smulticast
                \smultipath\s(?P<noneval>none)
                $""", re.VERBOSE,
            ),
            "setval": "ip multicast multipath none",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "multicast": {
                                "multipath": {
                                    "splitting_type": {
                                        "none": "{{ True if noneval is defined }}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip.multicast.multipath.splitting_type.legacy",
            "getval": re.compile(
                r"""
                \s+ip\smulticast
                \smultipath\s(?P<legacy_val>legacy)
                $""", re.VERBOSE,
            ),
            "setval": "ip multicast multipath legacy",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "multicast": {
                                "multipath": {
                                    "splitting_type": {
                                        "legacy": "{{ True if legacy_val is defined }}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip.multicast.multipath.splitting_type.nbm",
            "getval": re.compile(
                r"""
                \s+ip\smulticast
                \smultipath\s(?P<nbm_val>nbm)
                $""", re.VERBOSE,
            ),
            "setval": "ip multicast multipath nbm",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "multicast": {
                                "multipath": {
                                    "splitting_type": {
                                        "nbm": "{{ True if nbm_val is defined }}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip.multicast.multipath.splitting_type.sg_hash",
            "getval": re.compile(
                r"""
                \s+ip\smulticast
                \smultipath\s(?P<sg_hash_val>s-g-hash)
                $""", re.VERBOSE,
            ),
            "setval": "ip multicast multipath s-g-hash",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "multicast": {
                                "multipath": {
                                    "splitting_type": {
                                        "sg_hash": "{{ True if sg_hash_val is defined }}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip.multicast.multipath.splitting_type.sg_hash_next_hop",
            "getval": re.compile(
                r"""
                \s+ip\smulticast
                \smultipath
                \s(?P<sg_hash_nxt_val>s-g-hash\snext-hop-based)
                $""", re.VERBOSE,
            ),
            "setval": "ip multicast multipath s-g-hash next-hop-based",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "multicast": {
                                "multipath": {
                                    "splitting_type": {
                                        "sg_hash_next_hop": "{{ True if sg_hash_nxt_val is defined }}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip.multicast.rpf",
            "getval": re.compile(
                r"""
                \s+ip\smulticast
                \srpf\sselect
                \svrf\s(?P<vrf_val>\S+)
                \sgroup-list\s(?P<group_list>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "ip multicast rpf select vrf {{ vrf_name }} group-list {{ group_list_range }}",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "multicast": {
                                "rpf": [
                                    {
                                        "vrf_name": "{{ vrf_val }}",
                                        "group_list_range": "{{ group_list }}",
                                    },
                                ],
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip.name_server.use_vrf",
            "getval": re.compile(
                r"""
                \s+ip\sname-server
                \s(?P<source_addr>\S+)
                \suse-vrf\s(?P<vrf_name>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "ip name-server {{ ip.name_server.use_vrf.source_address }} use-vrf {{ ip.name_server.use_vrf.vrf }}",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "name_server": {
                                "use_vrf": {
                                    "source_address": "{{ source_addr }}",
                                    "vrf": "{{ vrf_name }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip.name_server.address_list",
            "getval": re.compile(
                r"""
                \s+ip\sname-server
                \s(?P<addr_list>.+$)
                $""", re.VERBOSE,
            ),
            "setval": "ip name-server {{ ip.name_server.address_list }}",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "name_server": {
                                "address_list": "{{ addr_list }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip.route",
            "getval": re.compile(
                r"""
                \s+ip\sroute
                \s(?P<src_val>\S+)
                \s(?P<dest_val>\S+)
                $""", re.VERBOSE,
            ),
            "setval": _tmplt_ip_route,
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "route": [
                                {
                                    "source": "{{ src_val }}",
                                    "destination": "{{ dest_val }}",
                                },
                            ],
                        },
                    },
                },
            },
        },
        {
            "name": "ip.route.tags",
            "getval": re.compile(
                r"""
                \s+ip\sroute
                \s(?P<src_val>\S+)
                \s(?P<dest_val>\S+)
                \stag\s(?P<tag_val>\d+)
                (\s(?P<route_pref_val>\d+))?
                $""", re.VERBOSE,
            ),
            "setval": _tmplt_ip_route,
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "route": [
                                {
                                    "source": "{{ src_val }}",
                                    "destination": "{{ dest_val }}",
                                    "tags": {
                                        "tag_value": "{{ tag_val }}",
                                        "route_pref": "{{ route_pref_val if route_pref_val is defined }}",
                                    },

                                },
                            ],
                        },
                    },
                },
            },
        },
        {
            "name": "ip.route.vrf",
            "getval": re.compile(
                r"""
                \s+ip\sroute
                \s(?P<src_val>\S+)
                \s(?P<dest_val>\S+)
                \svrf\s(?P<vrf_val>\S+)
                $""", re.VERBOSE,
            ),
            "setval": _tmplt_ip_route,
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "route": [
                                {
                                    "source": "{{ src_val }}",
                                    "destination": "{{ dest_val }}",
                                    "vrf": "{{ vrf_val }}",
                                },
                            ],
                        },
                    },
                },
            },
        },
        {
            "name": "ip.route.track",
            "getval": re.compile(
                r"""
                \s+ip\sroute
                \s(?P<src_val>\S+)
                \s(?P<dest_val>\S+)
                \strack\s(?P<track_val>\S+)
                $""", re.VERBOSE,
            ),
            "setval": _tmplt_ip_route,
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ip": {
                            "route": [
                                {
                                    "source": "{{ src_val }}",
                                    "destination": "{{ dest_val }}",
                                    "track": "{{ track_val }}",
                                },
                            ],
                        },
                    },
                },
            },
        },
        {
            "name": "vni",
            "getval": re.compile(
                r"""
                \s+vni\s(?P<vni_val>\d+)
                (\s(?P<l3_val>l3))?
                $""", re.VERBOSE,
            ),
            "setval": "vni {{ vni.vni_number }}{{ ' l3' if vni.layer_3 is defined }}",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "vni": {
                            "vni_number": "{{ vni_val }}",
                            "layer_3": "{{ True if l3_val is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "multicast.service_reflect",
            "getval": re.compile(
                r"""
                \s+multicast\sservice-reflect
                \sinterface\s(?P<serv_inter>\S+)
                \smap\sinterface\s(?P<map_inter>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "multicast service-reflect interface {{ service_interface }} map interface {{ map_to }}",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "multicast": {
                            "service_reflect": [
                                {
                                    "service_interface": "{{ serv_inter }}",
                                    "map_to": "{{ map_inter }}",
                                },
                            ],
                        },
                    },
                },
            },
        },
        {
            "name": "ipv6.mld_ssm_translate",
            "getval": re.compile(
                r"""
                \s+ipv6\smld
                (\s(?P<icmp_val>icmp))?
                \sssm-translate
                \s(?P<group_val>\S+)
                \s(?P<source_val>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "ipv6{{ ' icmp' if icmp is defined }} mld ssm-translate {{ group }} {{ source }}",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ipv6": {
                            "mld_ssm_translate": [
                                {
                                    "icmp": "{{ True if icmp_val is defined }}",
                                    "group": "{{ group_val }}",
                                    "source": "{{ source_val }}",
                                },
                            ],
                        },
                    },
                },
            },
        },
        {
            "name": "ipv6.multicast.group_range_prefix_list",
            "getval": re.compile(
                r"""
                \s+ipv6\smulticast
                \sgroup-range\sprefix-list
                \s(?P<prefix_lst>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "ipv6 multicast group-range prefix-list {{ ipv6.multicast.group_range_prefix_list }}",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ipv6": {
                            "multicast": {
                                "group_range_prefix_list": "{{ prefix_lst }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ipv6.multicast.multipath.resilient",
            "getval": re.compile(
                r"""
                \s+ipv6\smulticast
                \smultipath\s(?P<res_val>resilient)
                $""", re.VERBOSE,
            ),
            "setval": "ipv6 multicast multipath resilient",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ipv6": {
                            "multicast": {
                                "multipath": {
                                    "resilient": "{{ True if res_val is defined }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ipv6.multicast.multipath.splitting_type.none",
            "getval": re.compile(
                r"""
                \s+ipv6\smulticast
                \smultipath\s(?P<none_val>none)
                $""", re.VERBOSE,
            ),
            "setval": "ipv6 multicast multipath none",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ipv6": {
                            "multicast": {
                                "multipath": {
                                    "splitting_type": {
                                        "none": "{{ True if none_val is defined }}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ipv6.multicast.multipath.splitting_type.sg_hash",
            "getval": re.compile(
                r"""
                \s+ipv6\smulticast
                \smultipath\s(?P<sg_hash_val>s-g-hash)
                $""", re.VERBOSE,
            ),
            "setval": "ipv6 multicast multipath s-g-hash",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ipv6": {
                            "multicast": {
                                "multipath": {
                                    "splitting_type": {
                                        "sg_hash": "{{ True if sg_hash_val is defined }}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ipv6.multicast.multipath.splitting_type.sg_hash_next_hop",
            "getval": re.compile(
                r"""
                \s+ipv6\smulticast
                \smultipath\s(?P<sg_hash_next_hop_val>sg-nexthop-hash)
                $""", re.VERBOSE,
            ),
            "setval": "ipv6 multicast multipath sg-nexthop-hash",
            "result": {
                "vrfs": {
                    '{{ name }}': {
                        'name': '{{ name }}',
                        "ipv6": {
                            "multicast": {
                                "multipath": {
                                    "splitting_type": {
                                        "sg_hash_next_hop": "{{ True if sg_hash_next_hop_val is defined }}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
    ]
    # fmt: on



#
# -*- coding: utf-8 -*-
# Copyright 2020 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The iosxr_ospf_interfaces config file.
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to its desired end-state is
created.
"""


from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.resource_module import (
    ResourceModule,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    dict_merge,
)

from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.facts.facts import Facts
from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.rm_templates.ospf_interfaces import (
    Ospf_interfacesTemplate,
)


class Ospf_interfaces(ResourceModule):
    """
    The iosxr_ospf_interfaces config class
    """

    gather_subset = ["!all", "!min"]

    gather_network_resources = ["ospf_interfaces"]

    def __init__(self, module):
        super(Ospf_interfaces, self).__init__(
            empty_fact_val=[],
            facts_module=Facts(module),
            module=module,
            resource="ospf_interfaces",
            tmplt=Ospf_interfacesTemplate(),
        )
        self.parsers = [
            "authentication.message_digest",
            "authentication.null_auth",
            "authentication.message_digest.keychain",
            "authentication_key",
            "bfd.minimum_interval",
            "bfd.multiplier",
            "bfd.fast_detect.set",
            "bfd.fast_detect.disable",
            "bfd.fast_detect.strict_mode",
            "cost",
            "cost_fallback",
            "dead_interval",
            "demand_circuit",
            "flood_reduction",
            "hello_interval",
            "link_down.set",
            "link_down.disable",
            "message_digest_key",
            "mpls.set_ldp",
            "mpls.ldp_sync",
            "mpls.ldp_sync_disable",
            "mtu_ignore",
            "network",
            "packet_size",
            "passive",
            "prefix_suppression.disable",
            "prefix_suppression.secondary_address",
            "priority",
            "retransmit_interval",
            "security.ttl_hops",
            "security.ttl",
            "transmit_delay",
        ]

    def execute_module(self):
        """Execute the module
        :rtype: A dictionary
        :returns: The result from module execution
        """
        if self.state not in ["parsed", "gathered"]:
            self.generate_commands()
            self.run_commands()
        return self.result

    def generate_commands(self):
        """Generate configuration commands to send based on
        want, have and desired state.
        """
        wantd = {entry["name"]: entry for entry in self.want}
        haved = {entry["name"]: entry for entry in self.have}

        # turn all lists of dicts into dicts prior to merge
        for entry in wantd, haved:
            self._ospf_int_list_to_dict(entry)

        # if state is merged, merge want onto have and then compare
        if self.state == "merged":
            wantd = dict_merge(haved, wantd)

        # if state is deleted, empty out wantd and set haved to wantd
        if self.state == "deleted":
            haved = {k: v for k, v in haved.items() if k in wantd or not wantd}
            have_int = []
            if wantd == {}:
                for k, have in haved.items():
                    self._remove_ospf_int(have)

            for k, have in haved.items():
                if k in wantd:
                    have_int.append(k)
                    self._remove_ospf_int(have)
            wantd = {}

        # remove superfluous config for overridden and deleted
        if self.state == "overridden":
            for k, have in haved.items():
                if k not in wantd:
                    self._remove_ospf_int(have)
        if self.state != "deleted":
            for k, want in wantd.items():
                self._compare(want=want, have=haved.pop(k, {}))

    def _remove_ospf_int(self, entry):
        int_name = entry.get("name", {})
        int_addr = entry.get("address_family", {})
        for k, addr in int_addr.items():
            for key, value in addr["processes"].items():
                rem_entry = {
                    "name": int_name,
                    "afi": addr["afi"],
                    "process": value["process_id"],
                    "area": value["area"],
                }
            self.addcmd(rem_entry, "name", True)

    def _compare(self, want, have):
        """Leverages the base class `compare()` method and
        populates the list of commands to be run by comparing
        the `want` and `have` data with the `parsers` defined
        for the Ospf_interfaces network resource.
        """
        self._compare_interface(want=want, have=have)

        self._compare_addr_family(want=want, have=have)

    def _compare_interface(self, want, have):
        wdict = want.get("address_family", {})
        hdict = have.get("address_family", {})
        wname = want.get("name")
        hname = have.get("name")
        h_value = {}

        for key, w_value in wdict.items():
            if hdict and hdict.get(key):
                h_value = hdict[key]
            else:
                h_value = None
            w = {
                "name": wname,
                "type": w_value["afi"],
                "address_family": w_value,
            }
            if h_value is not None:
                h = {
                    "name": hname,
                    "type": h_value["afi"],
                    "address_family": h_value,
                }
            else:
                h = {}
            self.compare(parsers="name", want=w, have=h)

    def _compare_addr_family(self, want, have):
        wdict = want.get("address_family", {})
        hdict = have.get("address_family", {})
        wname = want.get("name")
        hname = have.get("name")
        # Fetch the area info as that would be common to all the attributes per interface
        for name, entry in wdict.items():
            w_process = {}
            h_process = {}
            for key, value in entry.items():
                if key == "processes":
                    for pname, pentry in value.items():
                        w_process = pentry
            for key, param in entry.items():
                w_addr = {"afi": name, key: param, "processes": w_process}
                h_addr = {}
                if hdict.get(name):
                    hdict_entry = hdict.get(name)
                    for item, value in hdict_entry.items():
                        if item == "processes":
                            for pname, pentry in value.items():
                                h_process = pentry
                    h_addr = {
                        "afi": name,
                        key: hdict[name].pop(key, {}),
                        "processes": h_process,
                    }
                w = {"name": wname, "address_family": w_addr}
                h = {"name": hname, "address_family": h_addr}
                self.compare(parsers=self.parsers, want=w, have=h)
        for name, entry in hdict.items():
            for key, param in entry.items():
                h_addr = {"afi": name, key: param}
                w_addr = {}
                w = {"name": wname, "address_family": w_addr}
                h = {"name": hname, "address_family": h_addr}
                self.compare(parsers=self.parsers, want=w, have=h)

    def _ospf_int_list_to_dict(self, entry):
        for name, family in entry.items():
            if "address_family" in family:
                family["address_family"] = {
                    entry["afi"]: entry for entry in family.get("address_family", [])
                }
                self._ospf_int_list_to_dict(family["address_family"])
        for name, ospf_processes in entry.items():
            if "processes" in ospf_processes:
                ospf_processes["processes"] = {
                    entry["process_id"]: entry for entry in ospf_processes.get("processes", [])
                }
                self._ospf_int_list_to_dict(ospf_processes["processes"])



# Copyright (c) 2020, Adam Migus <adam@migus.org>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
name: dsv
author: Adam Migus (@amigus) <adam@migus.org>
short_description: Get secrets from Thycotic DevOps Secrets Vault
version_added: 1.0.0
description:
  - Uses the Thycotic DevOps Secrets Vault Python SDK to get Secrets from a DSV O(tenant) using a O(client_id) and O(client_secret).
requirements:
  - python-dsv-sdk - https://pypi.org/project/python-dsv-sdk/
options:
  _terms:
    description: The path to the secret, for example V(/staging/servers/web1).
    required: true
  tenant:
    description: The first format parameter in the default O(url_template).
    type: string
    env:
      - name: DSV_TENANT
    ini:
      - section: dsv_lookup
        key: tenant
    required: true
  tld:
    default: com
    description: The top-level domain of the tenant; the second format parameter in the default O(url_template).
    type: string
    env:
      - name: DSV_TLD
    ini:
      - section: dsv_lookup
        key: tld
    required: false
  client_id:
    description: The client_id with which to request the Access Grant.
    type: string
    env:
      - name: DSV_CLIENT_ID
    ini:
      - section: dsv_lookup
        key: client_id
    required: true
  client_secret:
    description: The client secret associated with the specific O(client_id).
    type: string
    env:
      - name: DSV_CLIENT_SECRET
    ini:
      - section: dsv_lookup
        key: client_secret
    required: true
  url_template:
    default: https://{}.secretsvaultcloud.{}/v1
    description: The path to prepend to the base URL to form a valid REST API request.
    type: string
    env:
      - name: DSV_URL_TEMPLATE
    ini:
      - section: dsv_lookup
        key: url_template
    required: false
"""

RETURN = r"""
_list:
  description:
    - One or more JSON responses to C(GET /secrets/{path}).
    - See U(https://dsv.thycotic.com/api/index.html#operation/getSecret).
  type: list
  elements: dict
"""

EXAMPLES = r"""
- hosts: localhost
  vars:
    secret: "{{ lookup('community.general.dsv', '/test/secret') }}"
  tasks:
    - ansible.builtin.debug:
        msg: 'the password is {{ secret["data"]["password"] }}'
"""

from ansible.errors import AnsibleError, AnsibleOptionsError

from ansible_collections.community.general.plugins.plugin_utils._lookup import check_for_wrong_terms

sdk_is_missing = False

try:
    from thycotic.secrets.vault import (
        SecretsVault,
        SecretsVaultError,
    )
except ImportError:
    sdk_is_missing = True

from ansible.plugins.lookup import LookupBase
from ansible.utils.display import Display

display = Display()


class LookupModule(LookupBase):
    @staticmethod
    def Client(vault_parameters):
        try:
            vault = SecretsVault(**vault_parameters)
            return vault
        except TypeError as e:
            raise AnsibleError("python-dsv-sdk==0.0.1 must be installed to use this plugin") from e

    def run(self, terms, variables, **kwargs):
        if sdk_is_missing:
            raise AnsibleError("python-dsv-sdk==0.0.1 must be installed to use this plugin")

        self.set_options(var_options=variables, direct=kwargs)
        check_for_wrong_terms(self, direct=kwargs)

        vault = LookupModule.Client(
            {
                "tenant": self.get_option("tenant"),
                "client_id": self.get_option("client_id"),
                "client_secret": self.get_option("client_secret"),
                "tld": self.get_option("tld"),
                "url_template": self.get_option("url_template"),
            }
        )
        result = []

        for term in terms:
            display.debug(f"dsv_lookup term: {term}")
            try:
                path = term.lstrip("[/:]")

                if path == "":
                    raise AnsibleOptionsError(f"Invalid secret path: {term}")

                display.vvv(f"DevOps Secrets Vault GET /secrets/{path}")
                result.append(vault.get_secret_json(path))
            except SecretsVaultError as error:
                raise AnsibleError(f"DevOps Secrets Vault lookup failure: {error.message}") from error
        return result



# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Make coding more python3-ish
from __future__ import annotations

DOCUMENTATION = r"""
author: Unknown (!UNKNOWN)
name: syslog_json
type: notification
requirements:
  - whitelist in configuration
short_description: Sends JSON events to syslog
description:
  - This plugin logs ansible-playbook and ansible runs to a syslog server in JSON format.
options:
  server:
    description: Syslog server that receives the event.
    type: str
    env:
      - name: SYSLOG_SERVER
    default: localhost
    ini:
      - section: callback_syslog_json
        key: syslog_server
  port:
    description: Port on which the syslog server is listening.
    type: int
    env:
      - name: SYSLOG_PORT
    default: 514
    ini:
      - section: callback_syslog_json
        key: syslog_port
  facility:
    description: Syslog facility to log as.
    type: str
    env:
      - name: SYSLOG_FACILITY
    default: user
    ini:
      - section: callback_syslog_json
        key: syslog_facility
  setup:
    description: Log setup tasks.
    env:
      - name: ANSIBLE_SYSLOG_SETUP
    type: bool
    default: true
    ini:
      - section: callback_syslog_json
        key: syslog_setup
    version_added: 4.5.0
"""

import logging
import logging.handlers
import socket

from ansible.plugins.callback import CallbackBase


class CallbackModule(CallbackBase):
    """
    logs ansible-playbook and ansible runs to a syslog server in json format
    """

    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = "notification"
    CALLBACK_NAME = "community.general.syslog_json"
    CALLBACK_NEEDS_WHITELIST = True

    def __init__(self):
        super().__init__()

    def set_options(self, task_keys=None, var_options=None, direct=None):
        super().set_options(task_keys=task_keys, var_options=var_options, direct=direct)

        syslog_host = self.get_option("server")
        syslog_port = int(self.get_option("port"))
        syslog_facility = self.get_option("facility")

        self.logger = logging.getLogger("ansible logger")
        self.logger.setLevel(logging.DEBUG)

        self.handler = logging.handlers.SysLogHandler(address=(syslog_host, syslog_port), facility=syslog_facility)
        self.logger.addHandler(self.handler)
        self.hostname = socket.gethostname()

    def v2_runner_on_failed(self, result, ignore_errors=False):
        res = result._result
        host = result._host.get_name()
        self.logger.error(
            "%s ansible-command: task execution FAILED; host: %s; message: %s",
            self.hostname,
            host,
            self._dump_results(res),
        )

    def v2_runner_on_ok(self, result):
        res = result._result
        host = result._host.get_name()
        if result._task.action != "gather_facts" or self.get_option("setup"):
            self.logger.info(
                "%s ansible-command: task execution OK; host: %s; message: %s",
                self.hostname,
                host,
                self._dump_results(res),
            )

    def v2_runner_on_skipped(self, result):
        host = result._host.get_name()
        self.logger.info(
            "%s ansible-command: task execution SKIPPED; host: %s; message: %s", self.hostname, host, "skipped"
        )

    def v2_runner_on_unreachable(self, result):
        res = result._result
        host = result._host.get_name()
        self.logger.error(
            "%s ansible-command: task execution UNREACHABLE; host: %s; message: %s",
            self.hostname,
            host,
            self._dump_results(res),
        )

    def v2_runner_on_async_failed(self, result):
        res = result._result
        host = result._host.get_name()
        # jid = result._result.get("ansible_job_id")
        self.logger.error(
            "%s ansible-command: task execution FAILED; host: %s; message: %s",
            self.hostname,
            host,
            self._dump_results(res),
        )

    def v2_playbook_on_import_for_host(self, result, imported_file):
        host = result._host.get_name()
        self.logger.info(
            "%s ansible-command: playbook IMPORTED; host: %s; message: imported file %s",
            self.hostname,
            host,
            imported_file,
        )

    def v2_playbook_on_not_import_for_host(self, result, missing_file):
        host = result._host.get_name()
        self.logger.info(
            "%s ansible-command: playbook NOT IMPORTED; host: %s; message: missing file %s",
            self.hostname,
            host,
            missing_file,
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_campus_gateway_devices_uplinks_local_overrides_by_device_info
short_description: Information module for organizations _campus _gateway _devices
  _uplinks _local _overrides _by _device
description:
  - Get all organizations _campus _gateway _devices _uplinks _local _overrides _by
    _device.
  - Uplink overrides configured locally on Campus Gateway devices in an organization.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module_info
  - cisco.meraki.module_info_pagination
author: Francisco Munoz (@fmunoz)
options:
  headers:
    description: Additional headers.
    type: dict
  organizationId:
    description:
      - OrganizationId path parameter. Organization ID.
    type: str
  serials:
    description:
      - >
        Serials query parameter. A list of serial numbers. The returned devices will
        be filtered to only include these serials.
    elements: str
    type: list
  perPage:
    description:
      - PerPage query parameter. The number of entries per page returned. Acceptable
        range is 3 - 1000. Default is 1000.
    type: int
  startingAfter:
    description:
      - >
        StartingAfter query parameter. A token used by the server to indicate the
        start of the page. Often this is a timestamp or an ID but it is not limited
        to those. This parameter should not be defined by client applications. The
        link for the first, last, prev, or next page in the HTTP Link header should
        define it.
    type: str
  endingBefore:
    description:
      - >
        EndingBefore query parameter. A token used by the server to indicate the end
        of the page. Often this is a timestamp or an ID but it is not limited to those.
        This parameter should not be defined by client applications. The link for
        the first, last, prev, or next page in the HTTP Link header should define
        it.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for campusGateway getOrganizationCampusGatewayDevicesUplinksLocalOverridesByDevice
    description: Complete reference of the getOrganizationCampusGatewayDevicesUplinksLocalOverridesByDevice
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!get-organization-campus-gateway-devices-uplinks-local-overrides-by-device
notes:
  - SDK Method used are
    campus_gateway.CampusGateway.get_organization_campus_gateway_devices_uplinks_local_overrides_by_device,
  - Paths used are
    get /organizations/{organizationId}/campusGateway/devices/uplinks/localOverrides/byDevice,
"""

EXAMPLES = r"""
- name: Get all organizations _campus _gateway _devices _uplinks _local _overrides
    _by _device
  cisco.meraki.organizations_campus_gateway_devices_uplinks_local_overrides_by_device_info:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    serials: []
    perPage: 0
    startingAfter: string
    endingBefore: string
    organizationId: string
    total_pages: -1
    direction: next
  register: result
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "items": [
        {
          "serial": "string",
          "uplink": {
            "interface": "string",
            "vlan": 0,
            "addresses": [
              {
                "assignmentMode": "string",
                "protocol": "string",
                "gateway": "string",
                "subnetMask": "string",
                "address": "string"
              }
            ]
          },
          "nameservers": {
            "addresses": [
              "string"
            ]
          },
          "sgt": 0
        }
      ],
      "meta": {
        "counts": {
          "items": {
            "total": 0,
            "remaining": 0
          }
        }
      }
    }
"""



#!/usr/bin/python

# Copyright (c) 2016, Aleksei Kostiuk <unitoff@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: ipinfoio_facts
short_description: Retrieve IP geolocation facts of a host's IP address
description:
  - Gather IP geolocation facts of a host's IP address using ipinfo.io API.
author: "Aleksei Kostiuk (@akostyuk)"
extends_documentation_fragment:
  - community.general._attributes
  - community.general._attributes.facts
  - community.general._attributes.facts_module
options:
  timeout:
    description:
      - HTTP connection timeout in seconds.
    default: 10
    type: int
  http_agent:
    description:
      - Set http user agent.
    default: "ansible-ipinfoio-module/0.0.1"
    type: str
notes:
  - Check U(http://ipinfo.io/) for more information.
"""

EXAMPLES = r"""
# Retrieve geolocation data of a host's IP address
- name: Get IP geolocation data
  community.general.ipinfoio_facts:
"""

RETURN = r"""
ansible_facts:
  description: "Dictionary of IP geolocation facts for a host's IP address."
  returned: success
  type: dict
  contains:
    ip:
      description: "Public IP address of a host."
      type: str
      sample: "8.8.8.8"
    hostname:
      description: Domain name.
      type: str
      sample: "google-public-dns-a.google.com"
    country:
      description: ISO 3166-1 alpha-2 country code.
      type: str
      sample: "US"
    region:
      description: State or province name.
      type: str
      sample: "California"
    city:
      description: City name.
      type: str
      sample: "Mountain View"
    loc:
      description: Latitude and Longitude of the location.
      type: str
      sample: "37.3860,-122.0838"
    org:
      description: "Organization's name."
      type: str
      sample: "AS3356 Level 3 Communications, Inc."
    postal:
      description: Postal code.
      type: str
      sample: "94035"
"""
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.urls import fetch_url

USER_AGENT = "ansible-ipinfoio-module/0.0.1"


class IpinfoioFacts:
    def __init__(self, module):
        self.url = "https://ipinfo.io/json"
        self.timeout = module.params.get("timeout")
        self.module = module

    def get_geo_data(self):
        response, info = fetch_url(
            self.module,
            self.url,
            force=True,
            timeout=self.timeout,
        )
        if info["status"] != 200:
            self.module.fail_json(msg=f"Could not get {self.url} page, check for connectivity!")

        try:
            content = response.read()
            result = self.module.from_json(content)
        except ValueError:
            self.module.fail_json(msg=f"Failed to parse the ipinfo.io response: {self.url} {content}")
        else:
            return result


def main():
    module = AnsibleModule(
        argument_spec=dict(
            http_agent=dict(default=USER_AGENT),
            timeout=dict(type="int", default=10),
        ),
        supports_check_mode=True,
    )

    ipinfoio = IpinfoioFacts(module)
    ipinfoio_result = dict(changed=False, ansible_facts=ipinfoio.get_geo_data())
    module.exit_json(**ipinfoio_result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Mark Ciecior (@markciecior)
# Copyright: (c) 2024, Akini Ross <akinross@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "certified"}

DOCUMENTATION = r"""
---
module: aci_subject_label
short_description: Manage Subject Labels (vz:ConsSubjLbl and vz:ProvSubjLbl)
description:
- Manage Subject Labels on Cisco ACI fabrics.
options:
  tenant:
    description:
    - The name of the Tenant.
    type: str
    aliases: [ tenant_name ]
  l2out:
    description:
    - The name of the L2Out.
    type: str
    aliases: [ l2out_name ]
  l3out:
    description:
    - The name of the L3Out.
    type: str
    aliases: [ l3out_name ]
  external_epg:
    description:
    - The name of the External End Point Group.
    type: str
    aliases: [ extepg, extepg_name, external_epg_name ]
  contract:
    description:
    - The name of the Contract.
    type: str
    aliases: [ contract_name ]
  subject:
    description:
    - The name of the Subject.
    type: str
    aliases: [ subject_name ]
  ap:
    description:
    - The name of the Application Profile.
    type: str
    aliases: [ app_profile, app_profile_name, application_profile, application_profile_name]
  epg:
    description:
    - The name of the End Point Group.
    type: str
    aliases: [ epg_name ]
  esg:
    description:
    - The name of the Endpoint Security Group.
    type: str
    aliases: [ esg_name ]
  subject_label:
    description:
    - The name of the Subject Label.
    type: str
    aliases: [ subject_label_name, name, label ]
  subject_label_type:
    description:
    - Determines the type of the Subject Label.
    type: str
    required: true
    choices: [ consumer, provider ]
    aliases: [ type ]
  complement:
    description:
    - Whether complement is enabled on the Subject Label.
    - The APIC defaults to C(false) when unset during creation.
    type: bool
  tag:
    description:
    - The color of a policy label of the Subject Label.
    - The APIC defaults to C(yellow-green) when unset during creation.
    type: str
    choices:
    - alice_blue
    - antique_white
    - aqua
    - aquamarine
    - azure
    - beige
    - bisque
    - black
    - blanched_almond
    - blue
    - blue_violet
    - brown
    - burlywood
    - cadet_blue
    - chartreuse
    - chocolate
    - coral
    - cornflower_blue
    - cornsilk
    - crimson
    - cyan
    - dark_blue
    - dark_cyan
    - dark_goldenrod
    - dark_gray
    - dark_green
    - dark_khaki
    - dark_magenta
    - dark_olive_green
    - dark_orange
    - dark_orchid
    - dark_red
    - dark_salmon
    - dark_sea_green
    - dark_slate_blue
    - dark_slate_gray
    - dark_turquoise
    - dark_violet
    - deep_pink
    - deep_sky_blue
    - dim_gray
    - dodger_blue
    - fire_brick
    - floral_white
    - forest_green
    - fuchsia
    - gainsboro
    - ghost_white
    - gold
    - goldenrod
    - gray
    - green
    - green_yellow
    - honeydew
    - hot_pink
    - indian_red
    - indigo
    - ivory
    - khaki
    - lavender
    - lavender_blush
    - lawn_green
    - lemon_chiffon
    - light_blue
    - light_coral
    - light_cyan
    - light_goldenrod_yellow
    - light_gray
    - light_green
    - light_pink
    - light_salmon
    - light_sea_green
    - light_sky_blue
    - light_slate_gray
    - light_steel_blue
    - light_yellow
    - lime
    - lime_green
    - linen
    - magenta
    - maroon
    - medium_aquamarine
    - medium_blue
    - medium_orchid
    - medium_purple
    - medium_sea_green
    - medium_slate_blue
    - medium_spring_green
    - medium_turquoise
    - medium_violet_red
    - midnight_blue
    - mint_cream
    - misty_rose
    - moccasin
    - navajo_white
    - navy
    - old_lace
    - olive
    - olive_drab
    - orange
    - orange_red
    - orchid
    - pale_goldenrod
    - pale_green
    - pale_turquoise
    - pale_violet_red
    - papaya_whip
    - peachpuff
    - peru
    - pink
    - plum
    - powder_blue
    - purple
    - red
    - rosy_brown
    - royal_blue
    - saddle_brown
    - salmon
    - sandy_brown
    - sea_green
    - seashell
    - sienna
    - silver
    - sky_blue
    - slate_blue
    - slate_gray
    - snow
    - spring_green
    - steel_blue
    - tan
    - teal
    - thistle
    - tomato
    - turquoise
    - violet
    - wheat
    - white
    - white_smoke
    - yellow
    - yellow_green
  description:
    description:
    - The description for the Subject Label.
    type: str
    aliases: [ descr ]
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
  name_alias:
    description:
    - The alias for the current object. This relates to the nameAlias field in ACI.
    type: str
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation
- cisco.aci.owner

seealso:
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(vz:ConsSubjLbl) and (vz:ProvSubjLbl).
  link: https://developer.cisco.com/docs/apic-mim-ref/
author:
- Mark Ciecior (@markciecior)
- Akini Ross (@akinross)
"""

EXAMPLES = r"""
- name: Add a Subject Label on a Contract Subject
  cisco.aci.aci_subject_label:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    contract: web
    subject: web_subject
    subject_label: web_subject_label
    subject_type: consumer
    state: present
  delegate_to: localhost

- name: Add a Subject Label on a L2Out External EPG
  cisco.aci.aci_subject_label:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    l2out: l2out_name
    external_epg: external_epg_name
    subject_label: l2out_subject_label
    subject_type: consumer
    state: present
  delegate_to: localhost

- name: Add a Subject Label on a L3Out External EPG
  cisco.aci.aci_subject_label:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    l3out: l3out_name
    external_epg: external_epg_name
    subject_label: l3out_subject_label
    subject_type: consumer
    state: present
  delegate_to: localhost

- name: Add a Subject Label on a L3Out External EPG Contract
  cisco.aci.aci_subject_label:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    l3out: l3out_name
    external_epg: external_epg_name
    contract: web
    subject_label: l3out_subject_label
    subject_type: consumer
    state: present
  delegate_to: localhost

- name: Add a Subject Label on a ESG
  cisco.aci.aci_subject_label:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    ap: app_profile_name
    esg: esg_name
    subject_label: esg_subject_label
    subject_type: consumer
    state: present
  delegate_to: localhost

- name: Add a Subject Label on a EPG
  cisco.aci.aci_subject_label:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    ap: app_profile_name
    epg: epg_name
    subject_label: epg_subject_label
    subject_type: consumer
    state: present
  delegate_to: localhost

- name: Add a Subject Label on a EPG Contract
  cisco.aci.aci_subject_label:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    ap: app_profile_name
    epg: epg_name
    contract: web
    subject_label: epg_subject_label
    subject_type: consumer
    state: present
  delegate_to: localhost

- name: Query a Subject Label on a Contract Subject
  cisco.aci.aci_subject_label:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    contract: web
    subject: web_subject
    subject_label: web_subject_label
    subject_type: consumer
    state: query
  delegate_to: localhost
  register: query_result

- name: Query all Subject Labels
  cisco.aci.aci_subject_label:
    host: apic
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost
  register: query_result

- name: Remove a Subject Label on a Contract Subject
  cisco.aci.aci_subject_label:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    contract: web
    subject: web_subject
    subject_label: web_subject_label
    subject_type: consumer
    state: absent
  delegate_to: localhost
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec, aci_owner_spec
from ansible_collections.cisco.aci.plugins.module_utils.constants import ACI_CLASS_MAPPING, SUBJ_LABEL_MAPPING, SUBJ_LABEL_RN, POLICY_LABEL_COLORS


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(aci_owner_spec())
    argument_spec.update(
        tenant=dict(type="str", aliases=["tenant_name"]),
        l2out=dict(type="str", aliases=["l2out_name"]),
        l3out=dict(type="str", aliases=["l3out_name"]),
        external_epg=dict(type="str", aliases=["extepg", "extepg_name", "external_epg_name"]),
        contract=dict(type="str", aliases=["contract_name"]),
        subject=dict(type="str", aliases=["subject_name"]),
        ap=dict(type="str", aliases=["app_profile", "app_profile_name", "application_profile", "application_profile_name"]),
        epg=dict(type="str", aliases=["epg_name"]),
        esg=dict(type="str", aliases=["esg_name"]),
        complement=dict(type="bool"),
        description=dict(type="str", aliases=["descr"]),
        subject_label=dict(type="str", aliases=["subject_label_name", "name", "label"]),
        subject_label_type=dict(type="str", choices=["consumer", "provider"], aliases=["type"], required=True),
        tag=dict(type="str", choices=POLICY_LABEL_COLORS),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
        name_alias=dict(type="str"),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "present", ["tenant", "subject_label"]],
            ["state", "present", ["l2out", "l3out", "epg", "esg", "subject"], True],
            ["state", "absent", ["tenant", "subject_label"]],
            ["state", "absent", ["l2out", "l3out", "epg", "esg", "subject"], True],
        ],
        mutually_exclusive=[
            ["l2out", "l3out", "epg", "esg", "subject"],
            ["esg", "contract"],
            ["l2out", "contract"],
        ],
        required_by={
            "subject": ["contract"],
            "l2out": ["external_epg"],
            "l3out": ["external_epg"],
            "epg": ["ap"],
            "esg": ["ap"],
        },
    )

    aci = ACIModule(module)

    tenant = module.params.get("tenant")
    l2out = module.params.get("l2out")
    l3out = module.params.get("l3out")
    external_epg = module.params.get("external_epg")
    contract = module.params.get("contract")
    subject_label_type = module.params.get("subject_label_type")
    subject = module.params.get("subject")
    ap = module.params.get("ap")
    epg = module.params.get("epg")
    esg = module.params.get("esg")
    complement = aci.boolean(module.params.get("complement"))
    description = module.params.get("description")
    subject_label = module.params.get("subject_label")
    tag = module.params.get("tag")
    state = module.params.get("state")
    name_alias = module.params.get("name_alias")

    aci_class = SUBJ_LABEL_MAPPING.get(subject_label_type)
    aci_rn = SUBJ_LABEL_RN.get(subject_label_type) + subject_label if subject_label else None

    if contract:
        contract_rn = ACI_CLASS_MAPPING.get(subject_label_type).get("rn") + contract
        contract_class = ACI_CLASS_MAPPING.get(subject_label_type).get("class")

    root_class = dict(
        aci_class="fvTenant",
        aci_rn="tn-{0}".format(tenant),
        module_object=tenant,
        target_filter={"name": tenant},
    )
    subclass_1 = None
    subclass_2 = None
    subclass_3 = None
    subclass_4 = None
    if esg:
        subclass_1 = dict(
            aci_class="fvAp",
            aci_rn="ap-{0}".format(ap),
            module_object=ap,
            target_filter={"name": ap},
        )
        subclass_2 = dict(
            aci_class="fvESg",
            aci_rn="esg-{0}".format(esg),
            module_object=esg,
            target_filter={"name": esg},
        )
        subclass_3 = dict(
            aci_class=aci_class,
            aci_rn=aci_rn,
            module_object=subject_label,
            target_filter={"name": subject_label},
        )
    elif l2out:
        subclass_1 = dict(
            aci_class="l2extOut",
            aci_rn="l2out-{0}".format(l2out),
            module_object=l2out,
            target_filter={"name": l2out},
        )
        subclass_2 = dict(
            aci_class="l2extInstP",
            aci_rn="instP-{0}".format(external_epg),
            module_object=external_epg,
            target_filter={"name": external_epg},
        )
        subclass_3 = dict(
            aci_class=aci_class,
            aci_rn=aci_rn,
            module_object=subject_label,
            target_filter={"name": subject_label},
        )
    elif epg:
        subclass_1 = dict(
            aci_class="fvAp",
            aci_rn="ap-{0}".format(ap),
            module_object=ap,
            target_filter={"name": ap},
        )
        subclass_2 = dict(
            aci_class="fvAEPg",
            aci_rn="epg-{0}".format(epg),
            module_object=epg,
            target_filter={"name": epg},
        )
        if contract:
            subclass_3 = dict(
                aci_class=contract_class,
                aci_rn=contract_rn,
                module_object=contract,
                target_filter={"name": contract},
            )
            subclass_4 = dict(
                aci_class=aci_class,
                aci_rn=aci_rn,
                module_object=subject_label,
                target_filter={"name": subject_label},
            )
        else:
            subclass_3 = dict(
                aci_class=aci_class,
                aci_rn=aci_rn,
                module_object=subject_label,
                target_filter={"name": subject_label},
            )
    elif l3out:
        subclass_1 = subclass_1 = dict(
            aci_class="l3extOut",
            aci_rn="out-{0}".format(l3out),
            module_object=l3out,
            target_filter={"name": l3out},
        )
        subclass_2 = dict(
            aci_class="l3extInstP",
            aci_rn="instP-{0}".format(external_epg),
            module_object=external_epg,
            target_filter={"name": external_epg},
        )
        if contract:
            subclass_3 = dict(
                aci_class=contract_class,
                aci_rn=contract_rn,
                module_object=contract,
                target_filter={"name": contract},
            )
            subclass_4 = dict(
                aci_class=aci_class,
                aci_rn=aci_rn,
                module_object=subject_label,
                target_filter={"name": subject_label},
            )
        else:
            subclass_3 = dict(
                aci_class=aci_class,
                aci_rn=aci_rn,
                module_object=subject_label,
                target_filter={"name": subject_label},
            )
    elif subject:
        subclass_1 = dict(
            aci_class="vzBrCP",
            aci_rn="brc-{0}".format(contract),
            module_object=contract,
            target_filter={"name": contract},
        )
        subclass_2 = dict(
            aci_class="vzSubj",
            aci_rn="subj-{0}".format(subject),
            module_object=subject,
            target_filter={"name": subject},
        )
        subclass_3 = dict(
            aci_class=aci_class,
            aci_rn=aci_rn,
            module_object=subject_label,
            target_filter={"name": subject_label},
        )
    else:  # Query scenario without any filters forcing class query on the subject_label_class
        root_class = dict(
            aci_class=aci_class,
            aci_rn=aci_rn,
            module_object=subject_label,
            target_filter={"name": subject_label},
        )

    aci.construct_url(
        root_class=root_class,
        subclass_1=subclass_1,
        subclass_2=subclass_2,
        subclass_3=subclass_3,
        subclass_4=subclass_4,
    )

    aci.get_existing()

    if state == "present":
        aci.payload(
            aci_class=aci_class,
            class_config=dict(
                name=subject_label,
                descr=description,
                nameAlias=name_alias,
                isComplement=complement,
                tag=tag.replace("_", "-") if tag else None,
            ),
        )

        aci.get_diff(aci_class=aci_class)

        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2017 John Kwiatkoski (@JayKayy) <jkwiat40@gmail.com>
# Copyright (c) 2018 Alexander Bethke (@oolongbrothers) <oolongbrothers@gmx.net>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: flatpak
short_description: Manage flatpaks
description:
  - Allows users to add or remove flatpaks.
  - See the M(community.general.flatpak_remote) module for managing flatpak remotes.
author:
  - John Kwiatkoski (@JayKayy)
  - Alexander Bethke (@oolongbrothers)
requirements:
  - flatpak
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: partial
    details:
      - If O(state=latest), the module always returns RV(ignore:changed=true).
  diff_mode:
    support: none
options:
  executable:
    description:
      - The path to the C(flatpak) executable to use.
      - By default, this module looks for the C(flatpak) executable on the path.
    type: path
    default: flatpak
  method:
    description:
      - The installation method to use.
      - Defines if the C(flatpak) is supposed to be installed globally for the whole V(system) or only for the current V(user).
    type: str
    choices: [system, user]
    default: system
  name:
    description:
      - The name of the flatpak to manage. To operate on several packages this can accept a list of packages.
      - Should be specified as the unique reverse DNS name that identifies a flatpak (for example V(org.gnome.gedit)).
      - When supplying a reverse DNS name, you can use the O(remote) option to specify on what remote to look for the flatpak.
      - When used with O(state=present), O(name) can also be specified as a C(https://) or C(http://) URL to a C(flatpakref) file.
        However, it is recommended you use O(from_url) instead to get reliable idempotency.
        Passing URLs in O(name) will be deprecated in the future.
      - When used with O(state=absent) or O(state=latest), always specify the name in the reverse DNS format.
      - When supplying a URL with O(state=absent) or O(state=latest), the module tries to match the installed flatpak based
        on the name of the flatpakref to remove or update it. However, there is no guarantee that the names of the flatpakref
        file and the reverse DNS name of the installed flatpak do match.
    type: list
    elements: str
    required: true
  from_url:
    description:
      - A C(http://) or C(https://) URL pointing to a C(.flatpakref) file to install from.
      - When this option is set, O(name) must contain exactly one entry specifying the reverse DNS application ID of the
        flatpak (for example V(com.onepassword.OnePassword)). This is used to check whether the flatpak is already installed.
      - O(name) and O(from_url) cannot both contain URLs.
      - This option is recommended instead of passing a URL in O(name); passing URLs in O(name) will be deprecated in the future.
    type: str
    version_added: "12.6.0"
  no_dependencies:
    description:
      - If installing runtime dependencies should be omitted or not.
      - This parameter is primarily implemented for integration testing this module. There might however be some use cases
        where you would want to have this, like when you are packaging your own flatpaks.
    type: bool
    default: false
    version_added: 3.2.0
  remote:
    description:
      - The flatpak remote (repository) to install the flatpak from.
      - By default, V(flathub) is assumed, but you do need to add the flathub flatpak_remote before you can use this.
      - See the M(community.general.flatpak_remote) module for managing flatpak remotes.
    type: str
    default: flathub
  state:
    description:
      - Indicates the desired package state.
      - The value V(latest) is supported since community.general 8.6.0.
    choices: [absent, present, latest]
    type: str
    default: present
"""

EXAMPLES = r"""
- name: Install the spotify flatpak
  community.general.flatpak:
    name: com.spotify.Client
    from_url: https://s3.amazonaws.com/alexlarsson/spotify-repo/spotify.flatpakref
    state: present

- name: Install the gedit flatpak package without dependencies (not recommended)
  community.general.flatpak:
    name: org.gnome.gedit
    from_url: https://git.gnome.org/browse/gnome-apps-nightly/plain/gedit.flatpakref
    state: present
    no_dependencies: true

- name: Install the gedit package from flathub for current user
  community.general.flatpak:
    name: org.gnome.gedit
    state: present
    method: user

- name: Install the Gnome Calendar flatpak from the gnome remote system-wide
  community.general.flatpak:
    name: org.gnome.Calendar
    state: present
    remote: gnome

- name: Install GIMP using custom flatpak binary path
  community.general.flatpak:
    name: org.gimp.GIMP
    state: present
    executable: /usr/local/bin/flatpak-dev

- name: Install multiple packages
  community.general.flatpak:
    name:
      - org.gimp.GIMP
      - org.inkscape.Inkscape
      - org.mozilla.firefox

- name: Update the spotify flatpak
  community.general.flatpak:
    name: com.spotify.Client
    from_url: https://s3.amazonaws.com/alexlarsson/spotify-repo/spotify.flatpakref
    state: latest

- name: Update the gedit flatpak package without dependencies (not recommended)
  community.general.flatpak:
    name: org.gnome.gedit
    from_url: https://git.gnome.org/browse/gnome-apps-nightly/plain/gedit.flatpakref
    state: latest
    no_dependencies: true

- name: Update the gedit package from flathub for current user
  community.general.flatpak:
    name: org.gnome.gedit
    state: latest
    method: user

- name: Update the Gnome Calendar flatpak from the gnome remote system-wide
  community.general.flatpak:
    name: org.gnome.Calendar
    state: latest
    remote: gnome

- name: Update multiple packages
  community.general.flatpak:
    name:
      - org.gimp.GIMP
      - org.inkscape.Inkscape
      - org.mozilla.firefox
    state: latest

- name: Remove the gedit flatpak
  community.general.flatpak:
    name: org.gnome.gedit
    state: absent

- name: Remove multiple packages
  community.general.flatpak:
    name:
      - org.gimp.GIMP
      - org.inkscape.Inkscape
      - org.mozilla.firefox
    state: absent
"""

RETURN = r"""
command:
  description: The exact flatpak command that was executed.
  returned: When a flatpak command has been executed
  type: str
  sample: "/usr/bin/flatpak install --user --nontinteractive flathub org.gnome.Calculator"
"""

from re import match
from urllib.parse import urlparse

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._version import LooseVersion

OUTDATED_FLATPAK_VERSION_ERROR_MESSAGE = "Unknown option --columns=application"


def install_flat(module, binary, remote, names, method, no_dependencies, from_url=None):
    """Add new flatpaks."""
    global result  # pylint: disable=global-variable-not-assigned
    uri_names = []
    id_names = []
    for name in names:
        if name.startswith("http://") or name.startswith("https://"):
            uri_names.append(name)
        else:
            id_names.append(name)
    base_command = [binary, "install", f"--{method}"]
    flatpak_version = _flatpak_version(module, binary)
    if LooseVersion(flatpak_version) < LooseVersion("1.1.3"):
        base_command += ["-y"]
    else:
        base_command += ["--noninteractive"]
    if no_dependencies:
        base_command += ["--no-deps"]
    if from_url:
        command = base_command + ["--from", from_url]
        _flatpak_command(module, module.check_mode, command)
    else:
        if uri_names:
            command = base_command + uri_names
            _flatpak_command(module, module.check_mode, command)
        if id_names:
            command = base_command + [remote] + id_names
            _flatpak_command(module, module.check_mode, command)
    result["changed"] = True


def update_flat(module, binary, names, method, no_dependencies):
    """Update existing flatpaks."""
    global result  # pylint: disable=global-variable-not-assigned
    installed_flat_names = [_match_installed_flat_name(module, binary, name, method) for name in names]
    command = [binary, "update", f"--{method}"]
    flatpak_version = _flatpak_version(module, binary)
    if LooseVersion(flatpak_version) < LooseVersion("1.1.3"):
        command += ["-y"]
    else:
        command += ["--noninteractive"]
    if no_dependencies:
        command += ["--no-deps"]
    command += installed_flat_names
    stdout = _flatpak_command(module, module.check_mode, command)
    result["changed"] = (
        True if module.check_mode else (stdout.find("Nothing to do.") == -1 and stdout.find("Nothing to update.") == -1)
    )


def uninstall_flat(module, binary, names, method):
    """Remove existing flatpaks."""
    global result  # pylint: disable=global-variable-not-assigned
    installed_flat_names = [_match_installed_flat_name(module, binary, name, method) for name in names]
    command = [binary, "uninstall"]
    flatpak_version = _flatpak_version(module, binary)
    if LooseVersion(flatpak_version) < LooseVersion("1.1.3"):
        command += ["-y"]
    else:
        command += ["--noninteractive"]
    command += [f"--{method}"] + installed_flat_names
    _flatpak_command(module, module.check_mode, command)
    result["changed"] = True


def flatpak_exists(module, binary, names, method):
    """Check if the flatpaks are installed."""
    command = [binary, "list", f"--{method}"]
    output = _flatpak_command(module, False, command)
    installed = []
    not_installed = []
    for name in names:
        parsed_name = _parse_flatpak_name(name).lower()
        if parsed_name in output.lower():
            installed.append(name)
        else:
            not_installed.append(name)
    return installed, not_installed


def _match_installed_flat_name(module, binary, name, method):
    # This is a difficult function, since if the user supplies a flatpakref url,
    # we have to rely on a naming convention:
    # The flatpakref file name needs to match the flatpak name
    global result  # pylint: disable=global-variable-not-assigned
    parsed_name = _parse_flatpak_name(name)
    # Try running flatpak list with columns feature
    command = [binary, "list", f"--{method}", "--columns=application"]
    _flatpak_command(module, False, command, ignore_failure=True)
    if result["rc"] != 0 and OUTDATED_FLATPAK_VERSION_ERROR_MESSAGE in result["stderr"]:
        # Probably flatpak before 1.2
        matched_flatpak_name = _match_flat_using_flatpak_column_feature(module, binary, parsed_name, method)
    else:
        # Probably flatpak >= 1.2
        matched_flatpak_name = _match_flat_using_outdated_flatpak_format(module, binary, parsed_name, method)

    if matched_flatpak_name:
        return matched_flatpak_name
    else:
        result["msg"] = (
            "Flatpak removal failed: Could not match any installed flatpaks to "
            f"the name `{_parse_flatpak_name(name)}`. "
            "If you used a URL, try using the reverse DNS name of the flatpak"
        )
        module.fail_json(**result)


def _match_flat_using_outdated_flatpak_format(module, binary, parsed_name, method):
    global result  # pylint: disable=global-variable-not-assigned
    command = [binary, "list", f"--{method}", "--columns=application"]
    output = _flatpak_command(module, False, command)
    for row in output.split("\n"):
        if parsed_name.lower() == row.lower():
            return row


def _match_flat_using_flatpak_column_feature(module, binary, parsed_name, method):
    global result  # pylint: disable=global-variable-not-assigned
    command = [binary, "list", f"--{method}"]
    output = _flatpak_command(module, False, command)
    for row in output.split("\n"):
        if parsed_name.lower() in row.lower():
            return row.split()[0]


def _is_flatpak_id(part):
    # For guidelines on application IDs, refer to the following resources:
    # Flatpak:
    # https://docs.flatpak.org/en/latest/conventions.html#application-ids
    # Flathub:
    # https://docs.flathub.org/docs/for-app-authors/requirements#application-id
    return match(r"^[a-z]{2,}(\.\w+)+\.[\w-]+$", part)


def _parse_flatpak_name(name):
    if name.startswith("http://") or name.startswith("https://"):
        file_name = urlparse(name).path.split("/")[-1]
        file_name_without_extension = file_name.split(".")[0:-1]
        common_name = ".".join(file_name_without_extension)
    else:
        parts = name.split("/")
        for part in parts:
            if _is_flatpak_id(part):
                common_name = part
                break
        else:
            common_name = name
    return common_name


def _flatpak_version(module, binary):
    global result  # pylint: disable=global-variable-not-assigned
    command = [binary, "--version"]
    output = _flatpak_command(module, False, command)
    version_number = output.split()[1]
    return version_number


def _flatpak_command(module, noop, command, ignore_failure=False):
    global result  # pylint: disable=global-variable-not-assigned
    result["command"] = " ".join(command)
    if noop:
        result["rc"] = 0
        return ""

    result["rc"], result["stdout"], result["stderr"] = module.run_command(command, check_rc=not ignore_failure)
    return result["stdout"]


def main():
    # This module supports check mode
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(type="list", elements="str", required=True),
            from_url=dict(type="str"),
            remote=dict(type="str", default="flathub"),
            method=dict(type="str", default="system", choices=["user", "system"]),
            state=dict(type="str", default="present", choices=["absent", "present", "latest"]),
            no_dependencies=dict(type="bool", default=False),
            executable=dict(type="path", default="flatpak"),
        ),
        supports_check_mode=True,
    )

    name = module.params["name"]
    from_url = module.params["from_url"]
    state = module.params["state"]
    remote = module.params["remote"]
    no_dependencies = module.params["no_dependencies"]
    method = module.params["method"]
    executable = module.params["executable"]
    binary = module.get_bin_path(executable, None)

    global result
    result = dict(changed=False)

    # If the binary was not found, fail the operation
    if not binary:
        module.fail_json(msg=f"Executable '{executable}' was not found on the system.", **result)

    url_names = [n for n in name if n.startswith("http://") or n.startswith("https://")]
    if from_url is not None:
        if not (from_url.startswith("http://") or from_url.startswith("https://")):
            module.fail_json(msg="The 'from_url' parameter must be an http:// or https:// URL.", **result)
        if url_names:
            module.fail_json(msg="The 'name' and 'from_url' parameters cannot both contain URLs.", **result)
        if len(name) != 1:
            module.fail_json(msg="When 'from_url' is used, 'name' must contain exactly one entry.", **result)
    module.run_command_environ_update = dict(LANGUAGE="C", LC_ALL="C")

    installed, not_installed = flatpak_exists(module, binary, name, method)
    if state == "absent" and installed:
        uninstall_flat(module, binary, installed, method)
    else:
        if state == "latest" and installed:
            update_flat(module, binary, installed, method, no_dependencies)
        if state in ("present", "latest") and not_installed:
            install_flat(module, binary, remote, not_installed, method, no_dependencies, from_url)

    module.exit_json(**result)


if __name__ == "__main__":
    main()



# Copyright (c) 2022 Felix Fontein <felix@fontein.de>
# Copyright 2016 Red Hat | Ansible
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import abc
import os
import re
import shlex
import typing as t
from functools import partial

from ansible.module_utils.common.text.converters import to_text
from ansible.module_utils.common.text.formatters import human_to_bytes

from ansible_collections.community.docker.plugins.module_utils._api.utils.utils import (
    parse_env_file,
)
from ansible_collections.community.docker.plugins.module_utils._platform import (
    compare_platform_strings,
)
from ansible_collections.community.docker.plugins.module_utils._util import (
    clean_dict_booleans_for_docker_api,
    compare_generic,
    normalize_healthcheck,
    omit_none_from_dict,
    sanitize_labels,
)

if t.TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from ansible.module_utils.basic import AnsibleModule

    from ansible_collections.community.docker.plugins.module_utils._version import (
        LooseVersion,
    )

    ValueType = t.Literal["set", "list", "dict", "bool", "int", "float", "str"]
    AnsibleType = t.Literal["list", "dict", "bool", "int", "float", "str"]
    ComparisonMode = t.Literal["ignore", "strict", "allow_more_present"]
    ComparisonType = t.Literal["set", "set(dict)", "list", "dict", "value"]

Client = t.TypeVar("Client")


_DEFAULT_IP_REPLACEMENT_STRING = (
    "[[DEFAULT_IP:iewahhaeB4Sae6Aen8IeShairoh4zeph7xaekoh8Geingunaesaeweiy3ooleiwi]]"
)


_MOUNT_OPTION_TYPES = {
    "create_mountpoint": ("bind",),
    "labels": ("volume",),
    "no_copy": ("volume",),
    "non_recursive": ("bind",),
    "propagation": ("bind",),
    "read_only_force_recursive": ("bind",),
    "read_only_non_recursive": ("bind",),
    "subpath": ("volume", "image"),
    "tmpfs_size": ("tmpfs",),
    "tmpfs_mode": ("tmpfs",),
    "tmpfs_options": ("tmpfs",),
    "volume_driver": ("volume",),
    "volume_options": ("volume",),
}


def _get_ansible_type(
    value_type: ValueType,
) -> AnsibleType:
    if value_type == "set":
        return "list"
    if value_type not in ("list", "dict", "bool", "int", "float", "str"):
        raise ValueError(f'Invalid type "{value_type}"')
    return value_type


class Option:
    def __init__(
        self,
        name: str,
        *,
        value_type: ValueType,
        owner: OptionGroup,
        ansible_type: AnsibleType | None = None,
        elements: ValueType | None = None,
        ansible_elements: AnsibleType | None = None,
        ansible_suboptions: dict[str, t.Any] | None = None,
        ansible_aliases: Sequence[str] | None = None,
        ansible_choices: Sequence[str] | None = None,
        needs_no_suboptions: bool = False,
        default_comparison: ComparisonMode | None = None,
        not_a_container_option: bool = False,
        not_an_ansible_option: bool = False,
        copy_comparison_from: str | None = None,
        compare: Callable[[Option, t.Any, t.Any], bool] | None = None,
    ):
        self.name = name
        self.value_type = value_type
        self.ansible_type = ansible_type or _get_ansible_type(value_type)
        needs_elements = self.value_type in ("list", "set")
        needs_ansible_elements = self.ansible_type in ("list",)
        if elements is not None and not needs_elements:
            raise ValueError("elements only allowed for lists/sets")
        if elements is None and needs_elements:
            raise ValueError("elements required for lists/sets")
        if ansible_elements is not None and not needs_ansible_elements:
            raise ValueError("Ansible elements only allowed for Ansible lists")
        if (elements is None and ansible_elements is None) and needs_ansible_elements:
            raise ValueError("Ansible elements required for Ansible lists")
        self.elements = elements if needs_elements else None
        self.ansible_elements: AnsibleType | None = (
            (ansible_elements or _get_ansible_type(elements or "str"))
            if needs_ansible_elements
            else None
        )
        needs_suboptions = (
            self.ansible_type == "list" and self.ansible_elements == "dict"
        ) or (self.ansible_type == "dict")
        if ansible_suboptions is not None and not needs_suboptions:
            raise ValueError(
                "suboptions only allowed for Ansible lists with dicts, or Ansible dicts"
            )
        if (
            ansible_suboptions is None
            and needs_suboptions
            and not needs_no_suboptions
            and not not_an_ansible_option
        ):
            raise ValueError(
                "suboptions required for Ansible lists with dicts, or Ansible dicts"
            )
        self.ansible_suboptions = ansible_suboptions if needs_suboptions else None
        self.ansible_aliases = ansible_aliases or []
        self.ansible_choices = ansible_choices
        comparison_type: ComparisonType
        if self.value_type == "set" and self.elements == "dict":
            comparison_type = "set(dict)"
        elif self.value_type in ("set", "list", "dict"):
            comparison_type = self.value_type  # type: ignore
        else:
            comparison_type = "value"
        self.comparison_type = comparison_type
        if default_comparison is not None:
            self.comparison = default_comparison
        elif comparison_type in ("list", "value"):
            self.comparison = "strict"
        else:
            self.comparison = "allow_more_present"
        self.not_a_container_option = not_a_container_option
        self.not_an_ansible_option = not_an_ansible_option
        self.copy_comparison_from = copy_comparison_from
        self.compare = (
            (
                lambda param_value, container_value: compare(
                    self, param_value, container_value
                )
            )
            if compare
            else (
                lambda param_value, container_value: compare_generic(
                    param_value, container_value, self.comparison, self.comparison_type
                )
            )
        )


class OptionGroup:
    def __init__(
        self,
        *,
        preprocess: (
            Callable[[AnsibleModule, dict[str, t.Any]], dict[str, t.Any]] | None
        ) = None,
        ansible_mutually_exclusive: Sequence[Sequence[str]] | None = None,
        ansible_required_together: Sequence[Sequence[str]] | None = None,
        ansible_required_one_of: Sequence[Sequence[str]] | None = None,
        ansible_required_if: (
            Sequence[
                tuple[str, t.Any, Sequence[str]]
                | tuple[str, t.Any, Sequence[str], bool]
            ]
            | None
        ) = None,
        ansible_required_by: dict[str, Sequence[str]] | None = None,
    ) -> None:
        if preprocess is None:

            def preprocess(
                module: AnsibleModule, values: dict[str, t.Any]
            ) -> dict[str, t.Any]:
                return values

        self.preprocess = preprocess
        self.options: list[Option] = []
        self.all_options: list[Option] = []
        self.engines: dict[str, Engine] = {}
        self.ansible_mutually_exclusive = ansible_mutually_exclusive or []
        self.ansible_required_together = ansible_required_together or []
        self.ansible_required_one_of = ansible_required_one_of or []
        self.ansible_required_if = ansible_required_if or []
        self.ansible_required_by = ansible_required_by or {}
        self.argument_spec: dict[str, t.Any] = {}

    def add_option(self, name: str, **kwargs: t.Any) -> OptionGroup:
        option = Option(name, owner=self, **kwargs)
        if not option.not_a_container_option:
            self.options.append(option)
        self.all_options.append(option)
        if not option.not_an_ansible_option:
            ansible_option: dict[str, t.Any] = {
                "type": option.ansible_type,
            }
            if option.ansible_elements is not None:
                ansible_option["elements"] = option.ansible_elements
            if option.ansible_suboptions is not None:
                ansible_option["options"] = option.ansible_suboptions
            if option.ansible_aliases:
                ansible_option["aliases"] = option.ansible_aliases
            if option.ansible_choices is not None:
                ansible_option["choices"] = option.ansible_choices
            self.argument_spec[option.name] = ansible_option
        return self

    def supports_engine(self, engine_name: str) -> bool:
        return engine_name in self.engines

    def get_engine(self, engine_name: str) -> Engine:
        return self.engines[engine_name]

    def add_engine(self, engine_name: str, engine: Engine) -> OptionGroup:
        self.engines[engine_name] = engine
        return self


class Engine(t.Generic[Client]):
    min_api_version: str | None = None
    min_api_version_obj: LooseVersion | None = None
    extra_option_minimal_versions: dict[str, dict[str, t.Any]] | None = None

    @abc.abstractmethod
    def get_value(
        self,
        module: AnsibleModule,
        container: dict[str, t.Any],
        api_version: LooseVersion,
        options: list[Option],
        image: dict[str, t.Any] | None,
        host_info: dict[str, t.Any] | None,
    ) -> dict[str, t.Any]:
        pass

    def compare_value(
        self, option: Option, param_value: t.Any, container_value: t.Any
    ) -> bool:
        return option.compare(param_value, container_value)

    @abc.abstractmethod
    def set_value(
        self,
        module: AnsibleModule,
        data: dict[str, t.Any],
        api_version: LooseVersion,
        options: list[Option],
        values: dict[str, t.Any],
    ) -> None:
        pass

    @abc.abstractmethod
    def get_expected_values(
        self,
        module: AnsibleModule,
        client: Client,
        api_version: LooseVersion,
        options: list[Option],
        image: dict[str, t.Any] | None,
        values: dict[str, t.Any],
        host_info: dict[str, t.Any] | None,
    ) -> dict[str, t.Any]:
        pass

    @abc.abstractmethod
    def ignore_mismatching_result(
        self,
        module: AnsibleModule,
        client: Client,
        api_version: LooseVersion,
        option: Option,
        image: dict[str, t.Any] | None,
        container_value: t.Any,
        expected_value: t.Any,
        host_info: dict[str, t.Any] | None,
    ) -> bool:
        pass

    @abc.abstractmethod
    def preprocess_value(
        self,
        module: AnsibleModule,
        client: Client,
        api_version: LooseVersion,
        options: list[Option],
        values: dict[str, t.Any],
    ) -> dict[str, t.Any]:
        pass

    @abc.abstractmethod
    def update_value(
        self,
        module: AnsibleModule,
        data: dict[str, t.Any],
        api_version: LooseVersion,
        options: list[Option],
        values: dict[str, t.Any],
    ) -> None:
        pass

    @abc.abstractmethod
    def can_set_value(self, api_version: LooseVersion) -> bool:
        pass

    @abc.abstractmethod
    def can_update_value(self, api_version: LooseVersion) -> bool:
        pass

    @abc.abstractmethod
    def needs_container_image(self, values: dict[str, t.Any]) -> bool:
        pass

    @abc.abstractmethod
    def needs_host_info(self, values: dict[str, t.Any]) -> bool:
        pass


class EngineDriver(t.Generic[Client]):
    name: str

    @abc.abstractmethod
    def setup(
        self,
        argument_spec: dict[str, t.Any],
        mutually_exclusive: Sequence[Sequence[str]] | None = None,
        required_together: Sequence[Sequence[str]] | None = None,
        required_one_of: Sequence[Sequence[str]] | None = None,
        required_if: (
            Sequence[
                tuple[str, t.Any, Sequence[str]]
                | tuple[str, t.Any, Sequence[str], bool]
            ]
            | None
        ) = None,
        required_by: dict[str, Sequence[str]] | None = None,
    ) -> tuple[AnsibleModule, list[OptionGroup], Client]:
        pass

    @abc.abstractmethod
    def get_host_info(self, client: Client) -> dict[str, t.Any]:
        pass

    @abc.abstractmethod
    def get_api_version(self, client: Client) -> LooseVersion:
        pass

    @abc.abstractmethod
    def get_container_id(self, container: dict[str, t.Any]) -> str:
        pass

    @abc.abstractmethod
    def get_image_from_container(self, container: dict[str, t.Any]) -> str:
        pass

    @abc.abstractmethod
    def get_image_name_from_container(self, container: dict[str, t.Any]) -> str | None:
        pass

    @abc.abstractmethod
    def is_container_removing(self, container: dict[str, t.Any]) -> bool:
        pass

    @abc.abstractmethod
    def is_container_running(self, container: dict[str, t.Any]) -> bool:
        pass

    @abc.abstractmethod
    def is_container_paused(self, container: dict[str, t.Any]) -> bool:
        pass

    @abc.abstractmethod
    def inspect_container_by_name(
        self, client: Client, container_name: str
    ) -> dict[str, t.Any] | None:
        pass

    @abc.abstractmethod
    def inspect_container_by_id(
        self, client: Client, container_id: str
    ) -> dict[str, t.Any] | None:
        pass

    @abc.abstractmethod
    def inspect_image_by_id(
        self, client: Client, image_id: str
    ) -> dict[str, t.Any] | None:
        pass

    @abc.abstractmethod
    def inspect_image_by_name(
        self, client: Client, repository: str, tag: str
    ) -> dict[str, t.Any] | None:
        pass

    @abc.abstractmethod
    def pull_image(
        self,
        client: Client,
        repository: str,
        tag: str,
        image_platform: str | None = None,
    ) -> tuple[dict[str, t.Any] | None, bool]:
        pass

    @abc.abstractmethod
    def pause_container(self, client: Client, container_id: str) -> None:
        pass

    @abc.abstractmethod
    def unpause_container(self, client: Client, container_id: str) -> None:
        pass

    @abc.abstractmethod
    def disconnect_container_from_network(
        self, client: Client, container_id: str, network_id: str
    ) -> None:
        pass

    @abc.abstractmethod
    def connect_container_to_network(
        self,
        client: Client,
        container_id: str,
        network_id: str,
        parameters: dict[str, t.Any] | None = None,
    ) -> None:
        pass

    def create_container_supports_more_than_one_network(self, client: Client) -> bool:
        return False

    @abc.abstractmethod
    def create_container(
        self,
        client: Client,
        container_name: str,
        create_parameters: dict[str, t.Any],
        networks: dict[str, dict[str, t.Any]] | None = None,
    ) -> str:
        pass

    @abc.abstractmethod
    def start_container(self, client: Client, container_id: str) -> None:
        pass

    @abc.abstractmethod
    def wait_for_container(
        self, client: Client, container_id: str, timeout: int | float | None = None
    ) -> int | None:
        pass

    @abc.abstractmethod
    def get_container_output(
        self, client: Client, container_id: str
    ) -> tuple[bytes, t.Literal[True]] | tuple[str, t.Literal[False]]:
        pass

    @abc.abstractmethod
    def update_container(
        self, client: Client, container_id: str, update_parameters: dict[str, t.Any]
    ) -> None:
        pass

    @abc.abstractmethod
    def restart_container(
        self, client: Client, container_id: str, timeout: int | float | None = None
    ) -> None:
        pass

    @abc.abstractmethod
    def kill_container(
        self, client: Client, container_id: str, kill_signal: str | None = None
    ) -> None:
        pass

    @abc.abstractmethod
    def stop_container(
        self, client: Client, container_id: str, timeout: int | float | None = None
    ) -> None:
        pass

    @abc.abstractmethod
    def remove_container(
        self,
        client: Client,
        container_id: str,
        remove_volumes: bool = False,
        link: bool = False,
        force: bool = False,
    ) -> None:
        pass

    @abc.abstractmethod
    def run(self, runner: Callable[[], None], client: Client) -> None:
        pass


def _is_volume_permissions(mode: str) -> bool:
    for part in mode.split(","):
        if part not in (
            "rw",
            "ro",
            "z",
            "Z",
            "consistent",
            "delegated",
            "cached",
            "rprivate",
            "private",
            "rshared",
            "shared",
            "rslave",
            "slave",
            "nocopy",
        ):
            return False
    return True


def _parse_port_range(range_or_port: str, module: AnsibleModule) -> list[int]:
    """
    Parses a string containing either a single port or a range of ports.

    Returns a list of integers for each port in the list.
    """
    if "-" in range_or_port:
        try:
            start, end = [int(port) for port in range_or_port.split("-")]
        except ValueError:
            module.fail_json(msg=f'Invalid port range: "{range_or_port}"')
        if end < start:
            module.fail_json(msg=f'Invalid port range: "{range_or_port}"')
        return list(range(start, end + 1))
    try:
        return [int(range_or_port)]
    except ValueError:
        module.fail_json(msg=f'Invalid port: "{range_or_port}"')


def _split_colon_ipv6(text: str, module: AnsibleModule) -> list[str]:
    """
    Split string by ':', while keeping IPv6 addresses in square brackets in one component.
    """
    if "[" not in text:
        return text.split(":")
    start = 0
    result = []
    while start < len(text):
        i = text.find("[", start)
        if i < 0:
            result.extend(text[start:].split(":"))
            break
        j = text.find("]", i)
        if j < 0:
            module.fail_json(
                msg=f'Cannot find closing "]" in input "{text}" for opening "[" at index {i + 1}!'
            )
        result.extend(text[start:i].split(":"))
        k = text.find(":", j)
        if k < 0:
            result[-1] += text[i:]
            start = len(text)
        else:
            result[-1] += text[i:k]
            if k == len(text):
                result.append("")
                break
            start = k + 1
    return result


def _preprocess_command(
    module: AnsibleModule, values: dict[str, t.Any]
) -> dict[str, t.Any]:
    if "command" not in values:
        return values
    value = values["command"]
    if module.params["command_handling"] == "correct":
        if value is not None:
            if not isinstance(value, list):
                # convert from str to list
                value = shlex.split(to_text(value, errors="surrogate_or_strict"))
            value = [to_text(x, errors="surrogate_or_strict") for x in value]
    elif value:
        # convert from list to str
        if isinstance(value, list):
            value = shlex.split(
                " ".join([to_text(x, errors="surrogate_or_strict") for x in value])
            )
            value = [to_text(x, errors="surrogate_or_strict") for x in value]
        else:
            value = shlex.split(to_text(value, errors="surrogate_or_strict"))
            value = [to_text(x, errors="surrogate_or_strict") for x in value]
    else:
        return {}
    return {
        "command": value,
    }


def _preprocess_entrypoint(
    module: AnsibleModule, values: dict[str, t.Any]
) -> dict[str, t.Any]:
    if "entrypoint" not in values:
        return values
    value = values["entrypoint"]
    if module.params["command_handling"] == "correct":
        if value is not None:
            value = [to_text(x, errors="surrogate_or_strict") for x in value]
    elif value:
        # convert from list to str.
        value = shlex.split(
            " ".join([to_text(x, errors="surrogate_or_strict") for x in value])
        )
        value = [to_text(x, errors="surrogate_or_strict") for x in value]
    else:
        return {}
    return {
        "entrypoint": value,
    }


def _preprocess_env(
    module: AnsibleModule, values: dict[str, t.Any]
) -> dict[str, t.Any]:
    if not values:
        return {}
    final_env = {}
    if "env_file" in values:
        parsed_env_file = parse_env_file(values["env_file"])
        for name, value in parsed_env_file.items():
            final_env[name] = to_text(value, errors="surrogate_or_strict")
    if "env" in values:
        for name, value in values["env"].items():
            if not isinstance(value, str):
                module.fail_json(
                    msg="Non-string value found for env option. Ambiguous env options must be "
                    "wrapped in quotes to avoid them being interpreted when directly specified "
                    "in YAML, or explicitly converted to strings when the option is templated. "
                    f"Key: {name}"
                )
            final_env[name] = to_text(value, errors="surrogate_or_strict")
    formatted_env = []
    for key, value in final_env.items():
        formatted_env.append(f"{key}={value}")
    return {
        "env": formatted_env,
    }


def _preprocess_healthcheck(
    module: AnsibleModule, values: dict[str, t.Any]
) -> dict[str, t.Any]:
    if not values:
        return {}
    return {
        "healthcheck": normalize_healthcheck(
            values["healthcheck"], normalize_test=False
        ),
    }


def _preprocess_convert_to_bytes(
    module: AnsibleModule,
    values: dict[str, t.Any],
    name: str,
    unlimited_value: int | None = None,
) -> dict[str, t.Any]:
    if name not in values:
        return values
    try:
        value = values[name]
        if unlimited_value is not None and value in ("unlimited", str(unlimited_value)):
            value = unlimited_value
        else:
            value = human_to_bytes(value)
        values[name] = value
        return values
    except ValueError as exc:
        module.fail_json(msg=f"Failed to convert {name} to bytes: {exc}")


def _preprocess_mac_address(
    module: AnsibleModule, values: dict[str, t.Any]
) -> dict[str, t.Any]:
    if "mac_address" not in values:
        return values
    return {
        "mac_address": values["mac_address"].replace("-", ":"),
    }


def _preprocess_networks(
    module: AnsibleModule, values: dict[str, t.Any]
) -> dict[str, t.Any]:
    if (
        module.params["networks_cli_compatible"] is True
        and values.get("networks")
        and "network_mode" not in values
    ):
        # Same behavior as Docker CLI: if networks are specified, use the name of the first network as the value for network_mode
        # (assuming no explicit value is specified for network_mode)
        values["network_mode"] = values["networks"][0]["name"]

    if "networks" in values:
        for network in values["networks"]:
            if network["links"]:
                parsed_links = []
                for link in network["links"]:
                    parsed_link = link.split(":", 1)
                    if len(parsed_link) == 1:
                        parsed_link = (link, link)
                    parsed_links.append(tuple(parsed_link))
                network["links"] = parsed_links
            if network["mac_address"]:
                network["mac_address"] = network["mac_address"].replace("-", ":")

    return values


def _preprocess_sysctls(
    module: AnsibleModule, values: dict[str, t.Any]
) -> dict[str, t.Any]:
    if "sysctls" in values:
        for key, value in values["sysctls"].items():
            values["sysctls"][key] = to_text(value, errors="surrogate_or_strict")
    return values


def _preprocess_tmpfs(
    module: AnsibleModule, values: dict[str, t.Any]
) -> dict[str, t.Any]:
    if "tmpfs" not in values:
        return values
    result = {}
    for tmpfs_spec in values["tmpfs"]:
        split_spec = tmpfs_spec.split(":", 1)
        if len(split_spec) > 1:
            result[split_spec[0]] = split_spec[1]
        else:
            result[split_spec[0]] = ""
    return {"tmpfs": result}


def _preprocess_ulimits(
    module: AnsibleModule, values: dict[str, t.Any]
) -> dict[str, t.Any]:
    if "ulimits" not in values:
        return values
    result = []
    for limit in values["ulimits"]:
        limits = {}
        pieces = limit.split(":")
        if len(pieces) >= 2:
            limits["Name"] = pieces[0]
            limits["Soft"] = int(pieces[1])
            limits["Hard"] = int(pieces[1])
        if len(pieces) == 3:
            limits["Hard"] = int(pieces[2])
        result.append(limits)
    return {
        "ulimits": result,
    }


def _preprocess_mounts(
    module: AnsibleModule, values: dict[str, t.Any]
) -> dict[str, t.Any]:
    last: dict[str, str] = {}

    def check_collision(target: str, name: str) -> None:
        if target in last:
            if name == last[target]:
                module.fail_json(
                    msg=f'The mount point "{target}" appears twice in the {name} option'
                )
            else:
                module.fail_json(
                    msg=f'The mount point "{target}" appears both in the {name} and {last[target]} option'
                )
        last[target] = name

    if "mounts" in values:
        mounts = []
        for mount in values["mounts"]:
            target = mount["target"]
            mount_type = mount["type"]

            check_collision(target, "mounts")

            mount_dict = dict(mount)

            # Sanity checks
            if mount["source"] is None and mount_type not in (
                "tmpfs",
                "volume",
                "image",
                "cluster",
            ):
                module.fail_json(
                    msg=f'source must be specified for mount "{target}" of type "{mount_type}"'
                )
            for option, req_mount_types in _MOUNT_OPTION_TYPES.items():
                if mount[option] is not None and mount_type not in req_mount_types:
                    type_plural = "" if len(req_mount_types) == 1 else "s"
                    type_list = '", "'.join(req_mount_types)
                    module.fail_json(
                        msg=f'{option} cannot be specified for mount "{target}" of type "{mount_type}" (needs type{type_plural} "{type_list}")'
                    )

            # Streamline options
            volume_options = mount_dict.pop("volume_options")
            if mount_dict["volume_driver"] and volume_options:
                mount_dict["volume_options"] = clean_dict_booleans_for_docker_api(
                    volume_options
                )
            if mount_dict["labels"]:
                mount_dict["labels"] = clean_dict_booleans_for_docker_api(
                    mount_dict["labels"]
                )
            if mount_dict["tmpfs_size"] is not None:
                try:
                    mount_dict["tmpfs_size"] = human_to_bytes(mount_dict["tmpfs_size"])
                except ValueError as exc:
                    module.fail_json(
                        msg=f'Failed to convert tmpfs_size of mount "{target}" to bytes: {exc}'
                    )
            if mount_dict["tmpfs_mode"] is not None:
                try:
                    mount_dict["tmpfs_mode"] = int(mount_dict["tmpfs_mode"], 8)
                except ValueError:
                    module.fail_json(
                        msg=f'tmp_fs mode of mount "{target}" is not an octal string!'
                    )
            if mount_dict["tmpfs_options"]:
                opts = []
                for idx, opt in enumerate(mount_dict["tmpfs_options"]):
                    if len(opt) != 1:
                        module.fail_json(
                            msg=f'tmpfs_options[{idx + 1}] of mount "{target}" must be a one-element dictionary!'
                        )
                    k, v = list(opt.items())[0]
                    if not isinstance(k, str):
                        module.fail_json(
                            msg=f'key {k!r} in tmpfs_options[{idx + 1}] of mount "{target}" must be a string!'
                        )
                    if v is not None and not isinstance(v, str):
                        module.fail_json(
                            msg=f'value {v!r} in tmpfs_options[{idx + 1}] of mount "{target}" must be a string or null/none!'
                        )
                    opts.append([k, v] if v is not None else [k])
                mount_dict["tmpfs_options"] = opts

            # Add result to list
            mounts.append(omit_none_from_dict(mount_dict))
        values["mounts"] = mounts
    if "volumes" in values:
        new_vols = []
        for vol in values["volumes"]:
            parts = vol.split(":")
            if ":" in vol:
                if len(parts) == 3:
                    host, container, mode = parts
                    if not _is_volume_permissions(mode):
                        module.fail_json(msg=f"Found invalid volumes mode: {mode}")
                    if re.match(r"[.~]", host):
                        host = os.path.abspath(os.path.expanduser(host))
                    check_collision(container, "volumes")
                    new_vols.append(f"{host}:{container}:{mode}")
                    continue
                if (
                    len(parts) == 2
                    and not _is_volume_permissions(parts[1])
                    and re.match(r"[.~]", parts[0])
                ):
                    host = os.path.abspath(os.path.expanduser(parts[0]))
                    check_collision(parts[1], "volumes")
                    new_vols.append(f"{host}:{parts[1]}:rw")
                    continue
            check_collision(parts[min(1, len(parts) - 1)], "volumes")
            new_vols.append(vol)
        values["volumes"] = new_vols
        new_binds = []
        for vol in new_vols:
            host = None
            if ":" in vol:
                parts = vol.split(":")
                if len(parts) == 3:
                    host, container, mode = parts
                    if not _is_volume_permissions(mode):
                        module.fail_json(msg=f"Found invalid volumes mode: {mode}")
                elif len(parts) == 2:
                    if not _is_volume_permissions(parts[1]):
                        host, container, mode = parts + ["rw"]
            if host is not None:
                new_binds.append(f"{host}:{container}:{mode}")
        values["volume_binds"] = new_binds
    return values


def _preprocess_labels(
    module: AnsibleModule, values: dict[str, t.Any]
) -> dict[str, t.Any]:
    result = {}
    if "labels" in values:
        labels = values["labels"]
        if labels is not None:
            labels = dict(labels)
            sanitize_labels(labels, "labels", module=module)
        result["labels"] = labels
    return result


def _preprocess_log(
    module: AnsibleModule, values: dict[str, t.Any]
) -> dict[str, t.Any]:
    result: dict[str, t.Any] = {}
    if "log_driver" not in values:
        return result
    result["log_driver"] = values["log_driver"]
    if "log_options" in values:
        options: dict[str, str] = {}
        for k, v in values["log_options"].items():
            if not isinstance(v, str):
                value = to_text(v, errors="surrogate_or_strict")
                module.warn(
                    f"Non-string value found for log_options option '{k}'. The value is automatically converted to {value!r}. "
                    "If this is not correct, or you want to avoid such warnings, please quote the value,"
                    " or explicitly convert the values to strings when templating them."
                )
                v = value
            options[k] = v
        result["log_options"] = options
    return result


def _preprocess_ports(
    module: AnsibleModule, values: dict[str, t.Any]
) -> dict[str, t.Any]:
    if "published_ports" in values:
        if "all" in values["published_ports"]:
            module.fail_json(
                msg='Specifying "all" in published_ports is no longer allowed. Set publish_all_ports to "true" instead '
                "to randomly assign port mappings for those not specified by published_ports."
            )

        binds: dict[
            str | int,
            tuple[str]
            | tuple[str, str | int]
            | list[tuple[str] | tuple[str, str | int]],
        ] = {}
        for port in values["published_ports"]:
            parts = _split_colon_ipv6(
                to_text(port, errors="surrogate_or_strict"), module
            )
            container_port = parts[-1]
            protocol = ""
            if "/" in container_port:
                container_port, protocol = parts[-1].split("/")
            container_ports = _parse_port_range(container_port, module)

            p_len = len(parts)
            port_binds: Sequence[tuple[str] | tuple[str, str | int]]
            if p_len == 1:
                port_binds = len(container_ports) * [(_DEFAULT_IP_REPLACEMENT_STRING,)]
            elif p_len == 2:
                if len(container_ports) == 1:
                    port_binds = [(_DEFAULT_IP_REPLACEMENT_STRING, parts[0])]
                else:
                    port_binds = [
                        (_DEFAULT_IP_REPLACEMENT_STRING, port)
                        for port in _parse_port_range(parts[0], module)
                    ]
            elif p_len == 3:
                # We only allow IPv4 and IPv6 addresses for the bind address
                ipaddr = parts[0]
                if not re.match(
                    r"^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$", parts[0]
                ) and not re.match(r"^\[[0-9a-fA-F:]+(?:|%[^\]/]+)\]$", ipaddr):
                    module.fail_json(
                        msg="Bind addresses for published ports must be IPv4 or IPv6 addresses, not hostnames. "
                        f"Use the dig lookup to resolve hostnames. (Found hostname: {ipaddr})"
                    )
                if re.match(r"^\[[0-9a-fA-F:]+\]$", ipaddr):
                    ipaddr = ipaddr[1:-1]
                if parts[1]:
                    if len(container_ports) == 1:
                        port_binds = [(ipaddr, parts[1])]
                    else:
                        port_binds = [
                            (ipaddr, port)
                            for port in _parse_port_range(parts[1], module)
                        ]
                else:
                    port_binds = len(container_ports) * [(ipaddr,)]
            else:
                module.fail_json(
                    msg=f'Invalid port description "{port}" - expected 1 to 3 colon-separated parts, but got {p_len}. '
                    "Maybe you forgot to use square brackets ([...]) around an IPv6 address?"
                )

            for bind, container_port_val in zip(port_binds, container_ports):
                idx = (
                    f"{container_port_val}/{protocol}"
                    if protocol
                    else container_port_val
                )
                if idx in binds:
                    old_bind = binds[idx]
                    if isinstance(old_bind, list):
                        old_bind.append(bind)
                    else:
                        binds[idx] = [old_bind, bind]
                else:
                    binds[idx] = bind
        values["published_ports"] = binds

    exposed: set[tuple[int, str]] = set()
    if "exposed_ports" in values:
        for port in values["exposed_ports"]:
            port = to_text(port, errors="surrogate_or_strict").strip()
            protocol = "tcp"
            parts = port.split("/", maxsplit=1)
            if len(parts) == 2:
                port, protocol = parts
            parts = port.split("-", maxsplit=1)
            if len(parts) < 2:
                try:
                    exposed.add((int(port), protocol))
                except ValueError as e:
                    module.fail_json(msg=f"Cannot parse port {port!r}: {e}")
            else:
                try:
                    start_port = int(parts[0])
                    end_port = int(parts[1])
                    if start_port > end_port:
                        raise ValueError(
                            "start port must be smaller or equal to end port."
                        )
                except ValueError as e:
                    module.fail_json(msg=f"Cannot parse port range {port!r}: {e}")
                for port in range(start_port, end_port + 1):
                    exposed.add((port, protocol))
    if "published_ports" in values:
        # Any published port should also be exposed
        for publish_port in values["published_ports"]:
            if isinstance(publish_port, str) and "/" in publish_port:
                port, protocol = publish_port.split("/")
                port = int(port)
            else:
                protocol = "tcp"
                port = int(publish_port)
            exposed.add((port, protocol))
    values["ports"] = sorted(exposed)
    return values


def _compare_platform(
    option: Option, param_value: t.Any, container_value: t.Any
) -> bool:
    if option.comparison == "ignore":
        return True
    try:
        return compare_platform_strings(param_value, container_value)
    except ValueError:
        return param_value == container_value


OPTION_AUTO_REMOVE = OptionGroup().add_option("auto_remove", value_type="bool")

OPTION_BLKIO_WEIGHT = OptionGroup().add_option("blkio_weight", value_type="int")

OPTION_CAPABILITIES = OptionGroup().add_option(
    "capabilities", value_type="set", elements="str"
)

OPTION_CAP_DROP = OptionGroup().add_option("cap_drop", value_type="set", elements="str")

OPTION_CGROUP_NS_MODE = OptionGroup().add_option(
    "cgroupns_mode", value_type="str", ansible_choices=["private", "host"]
)

OPTION_CGROUP_PARENT = OptionGroup().add_option("cgroup_parent", value_type="str")

OPTION_COMMAND = OptionGroup(preprocess=_preprocess_command).add_option(
    "command", value_type="list", elements="str", ansible_type="raw"
)

OPTION_CPU_PERIOD = OptionGroup().add_option("cpu_period", value_type="int")

OPTION_CPU_QUOTA = OptionGroup().add_option("cpu_quota", value_type="int")

OPTION_CPUSET_CPUS = OptionGroup().add_option("cpuset_cpus", value_type="str")

OPTION_CPUSET_MEMS = OptionGroup().add_option("cpuset_mems", value_type="str")

OPTION_CPU_SHARES = OptionGroup().add_option("cpu_shares", value_type="int")

OPTION_ENTRYPOINT = OptionGroup(preprocess=_preprocess_entrypoint).add_option(
    "entrypoint", value_type="list", elements="str"
)

OPTION_CPUS = OptionGroup().add_option("cpus", value_type="int", ansible_type="float")

OPTION_DETACH_INTERACTIVE = (
    OptionGroup()
    .add_option("detach", value_type="bool")
    .add_option("interactive", value_type="bool")
)

OPTION_DEVICES = OptionGroup().add_option(
    "devices", value_type="set", elements="dict", ansible_elements="str"
)

OPTION_DEVICE_READ_BPS = OptionGroup().add_option(
    "device_read_bps",
    value_type="set",
    elements="dict",
    ansible_suboptions={
        "path": {"type": "str", "required": True},
        "rate": {"type": "str", "required": True},
    },
)

OPTION_DEVICE_WRITE_BPS = OptionGroup().add_option(
    "device_write_bps",
    value_type="set",
    elements="dict",
    ansible_suboptions={
        "path": {"type": "str", "required": True},
        "rate": {"type": "str", "required": True},
    },
)

OPTION_DEVICE_READ_IOPS = OptionGroup().add_option(
    "device_read_iops",
    value_type="set",
    elements="dict",
    ansible_suboptions={
        "path": {"type": "str", "required": True},
        "rate": {"type": "int", "required": True},
    },
)

OPTION_DEVICE_WRITE_IOPS = OptionGroup().add_option(
    "device_write_iops",
    value_type="set",
    elements="dict",
    ansible_suboptions={
        "path": {"type": "str", "required": True},
        "rate": {"type": "int", "required": True},
    },
)

OPTION_DEVICE_REQUESTS = OptionGroup().add_option(
    "device_requests",
    value_type="set",
    elements="dict",
    ansible_suboptions={
        "capabilities": {"type": "list", "elements": "list"},
        "count": {"type": "int"},
        "device_ids": {"type": "list", "elements": "str"},
        "driver": {"type": "str"},
        "options": {"type": "dict"},
    },
)

OPTION_DEVICE_CGROUP_RULES = OptionGroup().add_option(
    "device_cgroup_rules", value_type="list", elements="str"
)

OPTION_DNS_SERVERS = OptionGroup().add_option(
    "dns_servers", value_type="list", elements="str"
)

OPTION_DNS_OPTS = OptionGroup().add_option("dns_opts", value_type="set", elements="str")

OPTION_DNS_SEARCH_DOMAINS = OptionGroup().add_option(
    "dns_search_domains", value_type="list", elements="str"
)

OPTION_DOMAINNAME = OptionGroup().add_option("domainname", value_type="str")

OPTION_ENVIRONMENT = (
    OptionGroup(preprocess=_preprocess_env)
    .add_option(
        "env",
        value_type="set",
        ansible_type="dict",
        elements="str",
        needs_no_suboptions=True,
    )
    .add_option(
        "env_file",
        value_type="set",
        ansible_type="path",
        elements="str",
        not_a_container_option=True,
    )
)

OPTION_ETC_HOSTS = OptionGroup().add_option(
    "etc_hosts",
    value_type="set",
    ansible_type="dict",
    elements="str",
    needs_no_suboptions=True,
)

OPTION_GROUPS = OptionGroup().add_option("groups", value_type="set", elements="str")

OPTION_HEALTHCHECK = OptionGroup(preprocess=_preprocess_healthcheck).add_option(
    "healthcheck",
    value_type="dict",
    ansible_suboptions={
        "test": {"type": "raw"},
        "test_cli_compatible": {"type": "bool", "default": False},
        "interval": {"type": "str"},
        "timeout": {"type": "str"},
        "start_period": {"type": "str"},
        "start_interval": {"type": "str"},
        "retries": {"type": "int"},
    },
)

OPTION_HOSTNAME = OptionGroup().add_option("hostname", value_type="str")

OPTION_IMAGE = OptionGroup().add_option("image", value_type="str")

OPTION_INIT = OptionGroup().add_option("init", value_type="bool")

OPTION_IPC_MODE = OptionGroup().add_option("ipc_mode", value_type="str")

OPTION_KERNEL_MEMORY = OptionGroup(
    preprocess=partial(_preprocess_convert_to_bytes, name="kernel_memory")
).add_option("kernel_memory", value_type="int", ansible_type="str")

OPTION_LABELS = OptionGroup(preprocess=_preprocess_labels).add_option(
    "labels", value_type="dict", needs_no_suboptions=True
)

OPTION_LINKS = OptionGroup().add_option(
    "links", value_type="set", elements="list", ansible_elements="str"
)

OPTION_LOG_DRIVER_OPTIONS = (
    OptionGroup(
        preprocess=_preprocess_log, ansible_required_by={"log_options": ["log_driver"]}
    )
    .add_option("log_driver", value_type="str")
    .add_option(
        "log_options",
        value_type="dict",
        ansible_aliases=["log_opt"],
        needs_no_suboptions=True,
    )
)

OPTION_MAC_ADDRESS = OptionGroup(preprocess=_preprocess_mac_address).add_option(
    "mac_address", value_type="str"
)

OPTION_MEMORY = OptionGroup(
    preprocess=partial(_preprocess_convert_to_bytes, name="memory")
).add_option("memory", value_type="int", ansible_type="str")

OPTION_MEMORY_RESERVATION = OptionGroup(
    preprocess=partial(_preprocess_convert_to_bytes, name="memory_reservation")
).add_option("memory_reservation", value_type="int", ansible_type="str")

OPTION_MEMORY_SWAP = OptionGroup(
    preprocess=partial(
        _preprocess_convert_to_bytes, name="memory_swap", unlimited_value=-1
    )
).add_option("memory_swap", value_type="int", ansible_type="str")

OPTION_MEMORY_SWAPPINESS = OptionGroup().add_option(
    "memory_swappiness", value_type="int"
)

OPTION_STOP_TIMEOUT = OptionGroup().add_option(
    "stop_timeout", value_type="int", default_comparison="ignore"
)

OPTION_NETWORK = (
    OptionGroup(preprocess=_preprocess_networks)
    .add_option("network_mode", value_type="str")
    .add_option(
        "networks",
        value_type="set",
        elements="dict",
        ansible_suboptions={
            "name": {"type": "str", "required": True},
            "ipv4_address": {"type": "str"},
            "ipv6_address": {"type": "str"},
            "aliases": {"type": "list", "elements": "str"},
            "links": {"type": "list", "elements": "str"},
            "mac_address": {"type": "str"},
            "driver_opts": {"type": "dict"},
            "gw_priority": {"type": "int"},
        },
    )
)

OPTION_OOM_KILLER = OptionGroup().add_option("oom_killer", value_type="bool")

OPTION_OOM_SCORE_ADJ = OptionGroup().add_option("oom_score_adj", value_type="int")

OPTION_PID_MODE = OptionGroup().add_option("pid_mode", value_type="str")

OPTION_PIDS_LIMIT = OptionGroup().add_option("pids_limit", value_type="int")

OPTION_PLATFORM = OptionGroup().add_option(
    "platform", value_type="str", compare=_compare_platform
)

OPTION_PRIVILEGED = OptionGroup().add_option("privileged", value_type="bool")

OPTION_READ_ONLY = OptionGroup().add_option("read_only", value_type="bool")

OPTION_RESTART_POLICY = (
    OptionGroup(ansible_required_by={"restart_retries": ["restart_policy"]})
    .add_option(
        "restart_policy",
        value_type="str",
        ansible_choices=["no", "on-failure", "always", "unless-stopped"],
    )
    .add_option("restart_retries", value_type="int")
)

OPTION_RUNTIME = OptionGroup().add_option("runtime", value_type="str")

OPTION_SECURITY_OPTS = OptionGroup().add_option(
    "security_opts", value_type="set", elements="str"
)

OPTION_SHM_SIZE = OptionGroup(
    preprocess=partial(_preprocess_convert_to_bytes, name="shm_size")
).add_option("shm_size", value_type="int", ansible_type="str")

OPTION_STOP_SIGNAL = OptionGroup().add_option("stop_signal", value_type="str")

OPTION_STORAGE_OPTS = OptionGroup().add_option(
    "storage_opts", value_type="dict", needs_no_suboptions=True
)

OPTION_SYSCTLS = OptionGroup(preprocess=_preprocess_sysctls).add_option(
    "sysctls", value_type="dict", needs_no_suboptions=True
)

OPTION_TMPFS = OptionGroup(preprocess=_preprocess_tmpfs).add_option(
    "tmpfs", value_type="dict", ansible_type="list", ansible_elements="str"
)

OPTION_TTY = OptionGroup().add_option("tty", value_type="bool")

OPTION_ULIMITS = OptionGroup(preprocess=_preprocess_ulimits).add_option(
    "ulimits", value_type="set", elements="dict", ansible_elements="str"
)

OPTION_USER = OptionGroup().add_option("user", value_type="str")

OPTION_USERNS_MODE = OptionGroup().add_option("userns_mode", value_type="str")

OPTION_UTS = OptionGroup().add_option("uts", value_type="str")

OPTION_VOLUME_DRIVER = OptionGroup().add_option("volume_driver", value_type="str")

OPTION_VOLUMES_FROM = OptionGroup().add_option(
    "volumes_from", value_type="set", elements="str"
)

OPTION_WORKING_DIR = OptionGroup().add_option("working_dir", value_type="str")

OPTION_MOUNTS_VOLUMES = (
    OptionGroup(preprocess=_preprocess_mounts)
    .add_option(
        "mounts",
        value_type="set",
        elements="dict",
        ansible_suboptions={
            "target": {"type": "str", "required": True},
            "source": {"type": "str"},
            "type": {
                "type": "str",
                "choices": ["bind", "volume", "tmpfs", "npipe", "cluster", "image"],
                "default": "volume",
            },
            "read_only": {"type": "bool"},
            "consistency": {
                "type": "str",
                "choices": ["default", "consistent", "cached", "delegated"],
            },
            "propagation": {
                "type": "str",
                "choices": [
                    "private",
                    "rprivate",
                    "shared",
                    "rshared",
                    "slave",
                    "rslave",
                ],
            },
            "no_copy": {"type": "bool"},
            "labels": {"type": "dict"},
            "volume_driver": {"type": "str"},
            "volume_options": {"type": "dict"},
            "tmpfs_size": {"type": "str"},
            "tmpfs_mode": {"type": "str"},
            "non_recursive": {"type": "bool"},
            "create_mountpoint": {"type": "bool"},
            "read_only_non_recursive": {"type": "bool"},
            "read_only_force_recursive": {"type": "bool"},
            "subpath": {"type": "str"},
            "tmpfs_options": {"type": "list", "elements": "dict"},
        },
    )
    .add_option("volumes", value_type="set", elements="str")
    .add_option(
        "volume_binds",
        value_type="set",
        elements="str",
        not_an_ansible_option=True,
        copy_comparison_from="volumes",
    )
)

OPTION_PORTS = (
    OptionGroup(preprocess=_preprocess_ports)
    .add_option(
        "exposed_ports",
        value_type="set",
        elements="str",
        ansible_aliases=["exposed", "expose"],
    )
    .add_option("publish_all_ports", value_type="bool")
    .add_option(
        "published_ports",
        value_type="dict",
        ansible_type="list",
        ansible_elements="str",
        ansible_aliases=["ports"],
    )
    .add_option(
        "ports",
        value_type="set",
        elements="str",
        not_an_ansible_option=True,
        default_comparison="ignore",
    )
)

OPTIONS = [
    OPTION_AUTO_REMOVE,
    OPTION_BLKIO_WEIGHT,
    OPTION_CAPABILITIES,
    OPTION_CAP_DROP,
    OPTION_CGROUP_NS_MODE,
    OPTION_CGROUP_PARENT,
    OPTION_COMMAND,
    OPTION_CPU_PERIOD,
    OPTION_CPU_QUOTA,
    OPTION_CPUSET_CPUS,
    OPTION_CPUSET_MEMS,
    OPTION_CPU_SHARES,
    OPTION_ENTRYPOINT,
    OPTION_CPUS,
    OPTION_DETACH_INTERACTIVE,
    OPTION_DEVICES,
    OPTION_DEVICE_READ_BPS,
    OPTION_DEVICE_WRITE_BPS,
    OPTION_DEVICE_READ_IOPS,
    OPTION_DEVICE_WRITE_IOPS,
    OPTION_DEVICE_REQUESTS,
    OPTION_DEVICE_CGROUP_RULES,
    OPTION_DNS_SERVERS,
    OPTION_DNS_OPTS,
    OPTION_DNS_SEARCH_DOMAINS,
    OPTION_DOMAINNAME,
    OPTION_ENVIRONMENT,
    OPTION_ETC_HOSTS,
    OPTION_GROUPS,
    OPTION_HEALTHCHECK,
    OPTION_HOSTNAME,
    OPTION_IMAGE,
    OPTION_INIT,
    OPTION_IPC_MODE,
    OPTION_KERNEL_MEMORY,
    OPTION_LABELS,
    OPTION_LINKS,
    OPTION_LOG_DRIVER_OPTIONS,
    OPTION_MAC_ADDRESS,
    OPTION_MEMORY,
    OPTION_MEMORY_RESERVATION,
    OPTION_MEMORY_SWAP,
    OPTION_MEMORY_SWAPPINESS,
    OPTION_STOP_TIMEOUT,
    OPTION_NETWORK,
    OPTION_OOM_KILLER,
    OPTION_OOM_SCORE_ADJ,
    OPTION_PID_MODE,
    OPTION_PIDS_LIMIT,
    OPTION_PLATFORM,
    OPTION_PRIVILEGED,
    OPTION_READ_ONLY,
    OPTION_RESTART_POLICY,
    OPTION_RUNTIME,
    OPTION_SECURITY_OPTS,
    OPTION_SHM_SIZE,
    OPTION_STOP_SIGNAL,
    OPTION_STORAGE_OPTS,
    OPTION_SYSCTLS,
    OPTION_TMPFS,
    OPTION_TTY,
    OPTION_ULIMITS,
    OPTION_USER,
    OPTION_USERNS_MODE,
    OPTION_UTS,
    OPTION_VOLUME_DRIVER,
    OPTION_VOLUMES_FROM,
    OPTION_WORKING_DIR,
    OPTION_MOUNTS_VOLUMES,
    OPTION_PORTS,
]



# Copyright (c) 2020-2021, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import typing as t
from collections.abc import Callable

from ansible.errors import AnsibleFilterError
from ansible.module_utils.common.text.converters import to_text

from ansible_collections.community.dns.plugins.plugin_utils._public_suffix import (
    PUBLIC_SUFFIX_LIST,
)


def _remove_suffix(dns_name: str, suffix: str, keep_trailing_period: bool) -> str:
    suffix_len = len(suffix)
    if suffix_len and suffix_len < len(dns_name) and not keep_trailing_period:
        suffix_len += 1
    return dns_name[:-suffix_len] if suffix_len else dns_name


def get_registrable_domain(
    dns_name: t.Any,
    keep_unknown_suffix: t.Any = True,
    only_if_registerable: t.Any = True,
    normalize_result: t.Any = False,
    icann_only: t.Any = False,
) -> str:
    """Given DNS name, returns the registrable domain."""
    if not isinstance(dns_name, (str, bytes)):
        raise AnsibleFilterError(
            "Input for community.dns.get_registrable_domain must be a string"
        )
    for parameter, value in [
        ("keep_unknown_suffix", keep_unknown_suffix),
        ("only_if_registerable", only_if_registerable),
        ("normalize_result", normalize_result),
        ("icann_only", icann_only),
    ]:
        if not isinstance(value, bool):
            raise AnsibleFilterError(f"{parameter} must be a boolean, not {value!r}")
    return PUBLIC_SUFFIX_LIST.get_registrable_domain(
        to_text(dns_name),
        keep_unknown_suffix=keep_unknown_suffix,
        only_if_registerable=only_if_registerable,
        normalize_result=normalize_result,
        icann_only=icann_only,
    )


def get_public_suffix(
    dns_name: t.Any,
    keep_leading_period: t.Any = True,
    keep_unknown_suffix: t.Any = True,
    normalize_result: t.Any = False,
    icann_only: t.Any = False,
) -> str:
    """Given DNS name, returns the public suffix."""
    if not isinstance(dns_name, (str, bytes)):
        raise AnsibleFilterError(
            "Input for community.dns.get_registrable_domain must be a string"
        )
    for parameter, value in [
        ("keep_leading_period", keep_leading_period),
        ("keep_unknown_suffix", keep_unknown_suffix),
        ("normalize_result", normalize_result),
        ("icann_only", icann_only),
    ]:
        if not isinstance(value, bool):
            raise AnsibleFilterError(f"{parameter} must be a boolean, not {value!r}")
    suffix = PUBLIC_SUFFIX_LIST.get_suffix(
        to_text(dns_name),
        keep_unknown_suffix=keep_unknown_suffix,
        normalize_result=normalize_result,
        icann_only=icann_only,
    )
    if suffix and len(suffix) < len(dns_name) and keep_leading_period:
        suffix = "." + suffix
    return suffix


def remove_registrable_domain(
    dns_name: t.Any,
    keep_trailing_period: t.Any = False,
    keep_unknown_suffix: t.Any = True,
    only_if_registerable: t.Any = True,
    icann_only: t.Any = False,
) -> str:
    """Given DNS name, returns the part before the registrable_domain."""
    if not isinstance(dns_name, (str, bytes)):
        raise AnsibleFilterError(
            "Input for community.dns.remove_registrable_domain must be a string"
        )
    for parameter, value in [
        ("keep_trailing_period", keep_trailing_period),
        ("keep_unknown_suffix", keep_unknown_suffix),
        ("only_if_registerable", only_if_registerable),
        ("icann_only", icann_only),
    ]:
        if not isinstance(value, bool):
            raise AnsibleFilterError(f"{parameter} must be a boolean, not {value!r}")
    dns_name = to_text(dns_name)
    suffix = PUBLIC_SUFFIX_LIST.get_registrable_domain(
        dns_name,
        keep_unknown_suffix=keep_unknown_suffix,
        only_if_registerable=only_if_registerable,
        normalize_result=False,
        icann_only=icann_only,
    )
    return _remove_suffix(dns_name, suffix, keep_trailing_period)


def remove_public_suffix(
    dns_name: t.Any,
    keep_trailing_period: t.Any = False,
    keep_unknown_suffix: t.Any = True,
    icann_only: t.Any = False,
) -> str:
    """Given DNS name, returns the part before the public suffix."""
    if not isinstance(dns_name, (str, bytes)):
        raise AnsibleFilterError(
            "Input for community.dns.remove_public_suffix must be a string"
        )
    for parameter, value in [
        ("keep_trailing_period", keep_trailing_period),
        ("keep_unknown_suffix", keep_unknown_suffix),
        ("icann_only", icann_only),
    ]:
        if not isinstance(value, bool):
            raise AnsibleFilterError(f"{parameter} must be a boolean, not {value!r}")
    dns_name = to_text(dns_name)
    suffix = PUBLIC_SUFFIX_LIST.get_suffix(
        dns_name,
        keep_unknown_suffix=keep_unknown_suffix,
        normalize_result=False,
        icann_only=icann_only,
    )
    return _remove_suffix(dns_name, suffix, keep_trailing_period)


class FilterModule:
    """Ansible jinja2 filters"""

    def filters(self) -> dict[str, Callable]:
        return {
            "get_public_suffix": get_public_suffix,
            "get_registrable_domain": get_registrable_domain,
            "remove_public_suffix": remove_public_suffix,
            "remove_registrable_domain": remove_registrable_domain,
        }
